/* -*- C++ -*-
 * $Id: flchart.cc,v 1.12 2006/12/22 22:33:27 jeremy_henty Exp $
 */

#include "fltk.h"
#include "flchart.h"
#include "flenum.h"
#include "flcolor.h"

VALUE rb_cFLChart;

static FN_S_NEW(Chart);
WIDGET_FNS(Chart);
static FN_GETSET_BOOL(Chart,autosize);
static FN_CALL(Chart,clear);
static FN_GETSET_INT(Chart,maxsize);
static FN_GET_INT(Chart,size);
static FN_GETSET_ENUM(Chart,Fl_Font,textfont);
static FN_GETSET_INT(Chart,textsize);
static FN_GETSET_COLOR(Chart,textcolor);

static VALUE rb_FLChart_bounds(int argc, VALUE argv[], VALUE self)
{
  RBFLChart *widget = (RBFLChart*)rb_to_fl(self);
  VALUE arg1,arg2;
  double a,b;

  switch( rb_scan_args(argc, argv, "02", &arg1, &arg2) ){
  case 0:
    widget->bounds(&a,&b);
    return rb_assoc_new(rb_float_new(a), rb_float_new(b));
    break;
  case 1:
    Check_Type(arg1,T_ARRAY);
    arg2 = rb_ary_entry(arg1,1);
    arg1 = rb_ary_entry(arg1,0);
  case 2:
    arg1 = rb_Float(arg1);
    arg2 = rb_Float(arg2);
    widget->bounds(RFLOAT(arg1)->value, RFLOAT(arg2)->value);
    return Qnil;
  };
};

static VALUE rb_FLChart_add(int argc, VALUE argv[], VALUE self)
{
  RBFLChart *widget = (RBFLChart*)rb_to_fl(self);
  VALUE value;
  VALUE label;
  VALUE color;

  switch( rb_scan_args(argc,argv,"12",&value,&label,&color) ){
  case 1:
    value = rb_Float(value);
    widget->add(RFLOAT(value)->value);
    break;
  case 2:
    value = rb_Float(value);
    widget->add(RFLOAT(value)->value,STR2CPTR(label));
    break;
  case 3:
    value = rb_Float(value);
    widget->add(RFLOAT(value)->value,STR2CPTR(label),VALUE2COLOR(color));
    break;
  };
  return Qnil;
};

static VALUE rb_FLChart_insert(int argc, VALUE argv[], VALUE self)
{
  RBFLChart *widget = (RBFLChart*)rb_to_fl(self);
  VALUE pos;
  VALUE value;
  VALUE label;
  VALUE color;

  switch( rb_scan_args(argc,argv,"22",&pos,&value,&label,&color) ){
  case 2:
    value = rb_Float(value);
    widget->insert(NUM2INT(pos),RFLOAT(value)->value);
    break;
  case 3:
    value = rb_Float(value);
    widget->insert(NUM2INT(pos),RFLOAT(value)->value,STR2CPTR(label));
    break;
  case 4:
    value = rb_Float(value);
    widget->insert(NUM2INT(pos),RFLOAT(value)->value,STR2CPTR(label),VALUE2COLOR(color));
    break;
  };
  return Qnil;
};

static VALUE rb_FLChart_replace(int argc, VALUE argv[], VALUE self)
{
  RBFLChart *widget = (RBFLChart*)rb_to_fl(self);
  VALUE pos;
  VALUE value;
  VALUE label;
  VALUE color;

  switch( rb_scan_args(argc,argv,"22",&pos,&value,&label,&color) ){
  case 2:
    value = rb_Float(value);
    widget->replace(NUM2INT(pos),RFLOAT(value)->value);
    break;
  case 3:
    value = rb_Float(value);
    widget->replace(NUM2INT(pos),RFLOAT(value)->value,STR2CPTR(label));
    break;
  case 4:
    value = rb_Float(value);
    widget->replace(NUM2INT(pos),RFLOAT(value)->value,STR2CPTR(label),VALUE2COLOR(color));
    break;
  };
  return Qnil;
};

extern "C" void
Init_flchart(){
  rb_cFLChart = rb_define_class_under(rb_mFLTK, "Chart", rb_cFLWidget);
  DEF_S_METHOD(Chart,new,-1);
  DEF_WIDGET_METHODS(Chart);
  DEF_METHOD_EQ(Chart,autosize,-1);
  DEF_METHOD(Chart,clear,0);
  DEF_METHOD(Chart,add,-1);
  DEF_METHOD(Chart,insert,-1);
  DEF_METHOD(Chart,replace,-1);
  DEF_METHOD_EQ(Chart,maxsize,-1);
  DEF_METHOD(Chart,size,0);
  DEF_METHOD_EQ(Chart,bounds,-1);
  DEF_METHOD_EQ(Chart,textfont,-1);
  DEF_METHOD_EQ(Chart,textcolor,-1);
  DEF_METHOD_EQ(Chart,textsize,-1);

  DEF_MODULE_CONST_NUM(BAR_CHART);
  DEF_MODULE_CONST_NUM(FILL_CHART);
  DEF_MODULE_CONST_NUM(FILLED_CHART);
  DEF_MODULE_CONST_NUM(HORBAR_CHART);
  DEF_MODULE_CONST_NUM(LINE_CHART);
  DEF_MODULE_CONST_NUM(PIE_CHART);
  DEF_MODULE_CONST_NUM(SPECIALPIE_CHART);
  DEF_MODULE_CONST_NUM(SPIKE_CHART);

  DEF_MODULE_CONST_NUM(CHART_MAX);
  DEF_MODULE_CONST_NUM(CHART_LABEL_MAX);
};
