/* -*- C++ -*-
 * $Id: flhelp.cc,v 1.17 2006/12/31 11:30:36 jeremy_henty Exp $
 */

#include "fltk.h"

#ifdef HAVE_FL_FL_HELP_VIEW_H
#include "flhelp.h"
#include "flgroup.h"
#include "flcolor.h"

VALUE rb_cFLHelp_View;
VALUE rb_cFLHelp_Dialog;

static FN_S_NEW(Help_View);
HELP_FNS(Help_View);
static FN_INIT2(Help_View, Fl_Group::size, {
  rb_ivar_set(self, rb_id_ilink, Qnil);
});
static FN_GET_STR(Help_View,directory);
static FN_GET_STR(Help_View,filename);
static FN_SET_STR_COPY(Help_View,load);
static FN_GETSET_COLOR(Help_View,textcolor);
static FN_GETSET_INT(Help_View,textsize);
static FN_GETSET_INT(Help_View,textfont);
static FN_GETSET_INT(Help_View,topline);
static FN_GETSET_STR(Help_View,value);
static FN_GET_INT(Help_View,size);

static const char *rb_flhelp_link_i(Fl_Widget *w, const char *uri)
{
  VALUE obj = fl_to_rb(w);
  VALUE fn, r;
  if( obj != Qnil ){
    fn = rb_ivar_get(obj, rb_id_ilink);
    if( fn != Qnil ){
      r = rb_funcall(fn, rb_id_call, 2, obj, CPTR2STR(uri));
      return STR2CPTR(r);
    };
  };
  return uri;
};

static VALUE rb_FLHelp_View_link(int argc, VALUE argv[], VALUE self)
{
  RBFLHelp_View *widget = (RBFLHelp_View*)rb_to_fl(self);
  VALUE fn;

  switch( rb_scan_args(argc, argv, "01", &fn) ){
  case 0:
    fn = rb_block_proc();
    break;
  };
  rb_ivar_set(self, rb_id_ilink, fn);
  widget->link(rb_flhelp_link_i);
  return Qnil;
};

static VALUE rb_FLHelp_new(VALUE klass, Fl_Help_Dialog *dialog)
{
  VALUE obj = Data_Wrap_Struct(klass,0,0,dialog);
  rb_fltk_objmap_aset((Fl_Widget*)dialog, obj);
  return obj;
};

static VALUE rb_FLHelp_Dialog_s_new(VALUE klass)
{
  RBFLHelp_Dialog *dialog = new RBFLHelp_Dialog();
  VALUE obj;

  obj = rb_FLHelp_new(klass,dialog);
  rb_obj_call_init(obj,0,NULL);
  return obj;
}

static FN_CALL(Help_Dialog, hide);
static FN_CALL(Help_Dialog, show);
static FN_CALL(Help_Dialog, visible);
static FN_GETSET_INT(Help_Dialog, textsize);
//static FN_GETSET_INT(Help_Dialog, textfont);
static FN_GET_INT(Help_Dialog, x);
static FN_GET_INT(Help_Dialog, y);
static FN_GET_INT(Help_Dialog, w);
static FN_GET_INT(Help_Dialog, h);
static FN_SET_INT2(Help_Dialog, position);
static FN_SET_INT4(Help_Dialog, resize);

static VALUE rb_FLHelp_Dialog_topline(VALUE self,VALUE arg)
{
  RBFLHelp_Dialog *p = (RBFLHelp_Dialog*)rb_to_fl(self);

  if(TYPE(arg) == T_STRING)
    p->topline(STR2CSTR(arg));
  else
    p->topline(NUM2INT(arg));
  return self;
}

static VALUE rb_FLHelp_Dialog_load(VALUE self,VALUE filename)
{
  RBFLHelp_Dialog *p = (RBFLHelp_Dialog*)rb_to_fl(self);

  p->load(STR2CSTR(filename));

  return self;
}

static VALUE rb_FLHelp_Dialog_destroy(VALUE self)
{
  RBFLHelp_Dialog *p = (RBFLHelp_Dialog*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)p);
  return Qnil;
};

extern "C" void Init_flhelp()
{
  rb_cFLHelp_View   = rb_define_class_under(rb_mFLTK,"HelpView",rb_cFLGroup);
  DEF_S_METHOD(Help_View,new,-1);
  DEF_HELP_METHODS(Help_View);
  DEF_INIT_METHOD(Help_View);
  DEF_METHOD(Help_View,directory,0);
  DEF_METHOD(Help_View,filename,0);
  DEF_METHOD(Help_View,load,1);
  DEF_METHOD(Help_View,size,0);
  DEF_METHOD_EQ(Help_View,textcolor,-1);
  DEF_METHOD_EQ(Help_View,textsize,-1);
  DEF_METHOD_EQ(Help_View,textfont,-1);
  DEF_METHOD_EQ(Help_View,topline,-1);
  DEF_METHOD_EQ(Help_View,value,-1);
  DEF_METHOD_EQ(Help_View,link,-1);

  rb_cFLHelp_Dialog = rb_define_class_under(rb_mFLTK,"HelpDialog",rb_cObject);
  DEF_S_METHOD(Help_Dialog,new,0);
  DEF_METHOD(Help_Dialog,load,1);
  DEF_METHOD_EQ(Help_Dialog,topline,1);
  DEF_METHOD(Help_Dialog,hide,0);
  DEF_METHOD(Help_Dialog,show,0);
  DEF_METHOD_Q(Help_Dialog,visible,0);
  DEF_METHOD(Help_Dialog,destroy,0);
  DEF_METHOD_EQ(Help_Dialog, textsize,-1);
  //DEF_METHOD_EQ(Help_Dialog,textfont,-1);
  DEF_METHOD(Help_Dialog,x,0);
  DEF_METHOD(Help_Dialog,y,0);
  DEF_METHOD(Help_Dialog,w,0);
  DEF_METHOD(Help_Dialog,h,0);
  DEF_METHOD(Help_Dialog,position,2);
  DEF_METHOD(Help_Dialog,resize,4);
}
#endif /* HAVE_FL_FL_HELP_VIEW_H */
