/* -*- C++ -*-
 * $Id: flimage.cc,v 1.16 2006/12/29 05:11:34 jeremy_henty Exp $
 */

#include "fltk.h"
#include "flimage.h"
#include "flcolor.h"

VALUE rb_cFLImage;
VALUE rb_cFLRGB_Image;
VALUE rb_cFLPNG_Image;
VALUE rb_cFLJPEG_Image;
VALUE rb_cFLBMP_Image;
VALUE rb_cFLTiled_Image;
VALUE rb_cFLShared_Image;

static void rb_FLImage_free(Fl_Image *image)
{
};

static VALUE rb_FLImage_new(VALUE klass, Fl_Image *image)
{
  VALUE obj;
  obj = Data_Wrap_Struct(klass, 0, rb_FLImage_free, image);
  rb_fltk_objmap_aset((Fl_Widget*)image, obj);
  return obj;
};

static VALUE rb_FLImage_initialize(int argc, VALUE argv[], VALUE self)
{
  return Qnil;
};

#if FLTK_VERSION < FLTK_VERSION_CODE(1,1,0)
static VALUE rb_FLImage_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, xbm, str, w, h, d, ld;
  RBFLImage *image;
  int ary_len, str_len;
  uchar *data;
  uchar *rb_fltk_create_image(VALUE);

  rb_scan_args(argc, argv, "32", &xbm,&w,&h, &d, &ld);

  if( TYPE(xbm) == T_ARRAY ){
    data = rb_fltk_create_image(xbm);
  }
  else{
    Check_Type(xbm, T_STRING);
    data = ALLOC_N(uchar,RSTRING(xbm)->len);
    memcpy(data, RSTRING(xbm)->ptr, RSTRING(xbm)->len);
  }

  switch( argc ){
  case 3:
    image = new RBFLImage(data,NUM2INT(w),NUM2INT(h));
    break;
  case 4:
    image = new RBFLImage(data,NUM2INT(w),NUM2INT(h),NUM2INT(d));
    break;
  case 5:
    image = new RBFLImage(data,NUM2INT(w),NUM2INT(h),NUM2INT(d), NUM2INT(ld));
    break;
  }
  obj = rb_FLImage_new(klass, image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

static VALUE rb_FLImage_destroy(VALUE self)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)image);
  if( image && image->array ){
    free((char*)(image->array));
    delete image;
  };

  return Qnil;
};

#else /* if FLTK_VERSION < FLTK_VERSION_CODE(...) */

static VALUE rb_FLImage_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, w, h, d;
  RBFLImage *image;
  uchar *rb_fltk_create_image(VALUE);

  rb_scan_args(argc, argv, "30", &w,&h,&d);

  image = new RBFLImage(NUM2INT(w),NUM2INT(h),NUM2INT(d));
  obj = rb_FLImage_new(klass,image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

static VALUE rb_FLRGB_Image_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, xbm, w, h, d, ld;
  RBFLRGB_Image *image;
  uchar *data;
  uchar *rb_fltk_create_image(VALUE);

  rb_scan_args(argc, argv, "32", &xbm,&w,&h, &d, &ld);

  if( TYPE(xbm) == T_ARRAY ){
    data = rb_fltk_create_image(xbm);
  }
  else{
    Check_Type(xbm, T_STRING);
    data = ALLOC_N(uchar,RSTRING(xbm)->len);
    memcpy(data, RSTRING(xbm)->ptr, RSTRING(xbm)->len);
  }

  switch( argc ){
  case 3:
    image = new RBFLRGB_Image(data,NUM2INT(w),NUM2INT(h));
    break;
  case 4:
    image = new RBFLRGB_Image(data,NUM2INT(w),NUM2INT(h),NUM2INT(d));
    break;
  case 5:
    image = new RBFLRGB_Image(data,NUM2INT(w),NUM2INT(h),NUM2INT(d), NUM2INT(ld));
    break;
  }
  obj = rb_FLImage_new(klass,image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};


static VALUE rb_FLPNG_Image_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, filename;
  RBFLPNG_Image *image;

  rb_scan_args(argc, argv, "1", &filename);

  image = new RBFLPNG_Image(STR2CSTR(filename));
  obj = rb_FLImage_new(klass,image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

static VALUE rb_FLJPEG_Image_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, filename;
  RBFLJPEG_Image *image;

  rb_scan_args(argc, argv, "1", &filename);

  image = new RBFLJPEG_Image(STR2CSTR(filename));
  obj = rb_FLImage_new(klass,image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

static VALUE rb_FLBMP_Image_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, filename;
  RBFLBMP_Image *image;

  rb_scan_args(argc, argv, "1", &filename);

  image = new RBFLBMP_Image(STR2CSTR(filename));
  obj = rb_FLImage_new(klass,image);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

static VALUE rb_FLImage_destroy(VALUE self)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)image);
  if( image ){
    delete image;
  };

  return Qnil;
};

static VALUE rb_FLRGB_Image_destroy(VALUE self)
{
  RBFLRGB_Image *image = (RBFLRGB_Image*)rb_to_fl(self);

  rb_fltk_objmap_delete((Fl_Widget*)image);
  if( image && image->array ){
    free((char*)(image->array));
    delete image;
  };

  return Qnil;
};

#endif

static VALUE rb_FLImage_label(VALUE self, VALUE vwidget)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);
  RBFLWidget *widget = (RBFLWidget*)rb_to_fl(vwidget);

  image->label(widget);

  return Qnil;
};

static VALUE rb_FLImage_draw(int argc, VALUE argv[], VALUE self)
{
  RBFLImage *image = (RBFLImage*)rb_to_fl(self);
  VALUE arg1, arg2, arg3, arg4, arg5, arg6;
  int x, y, w, h, cx, cy;

  switch( rb_scan_args(argc, argv, "24", &arg1,&arg2,&arg3,&arg4,&arg5,&arg6) ){
  case 2:
    x = NUM2INT(arg1);
    y = NUM2INT(arg2);
    image->draw(x,y);
    break;
  case 4:
    x = NUM2INT(arg1);
    y = NUM2INT(arg2);
    w = NUM2INT(arg3);
    h = NUM2INT(arg4);
    image->draw(x,y,w,h);
    break;
  case 6:
    x = NUM2INT(arg1);
    y = NUM2INT(arg2);
    w = NUM2INT(arg3);
    h = NUM2INT(arg4);
    cx = NUM2INT(arg5);
    cy = NUM2INT(arg6);
    image->draw(x,y,w,h,cx,cy);
    break;
  default:
    rb_raise(rb_eArgError, "too few/many arguments");
  };

  return Qnil;
};


#if FLTK_VERSION >= FLTK_VERSION_CODE(1,1,0)
static VALUE rb_FLTiled_Image_s_new(int argc, VALUE argv[], VALUE klass)
{
  VALUE obj, img, w, h;
  RBFLTiled_Image *flimg;

  switch( rb_scan_args(argc, argv, "12", &img, &w, &h) ){
  case 1:
    flimg = new RBFLTiled_Image((RBFLImage*)rb_to_fl(img));
    break;
  case 3:
    flimg = new RBFLTiled_Image((RBFLImage*)rb_to_fl(img),
				(int)NUM2INT(w), (int)NUM2INT(h));
    break;
  default:
    rb_raise(rb_eArgError, "too few/many arguments");
  };

  obj = rb_FLImage_new(klass, flimg);

  rb_obj_call_init(obj, argc, argv);

  return obj;
};

static VALUE rb_FLTiled_Image_copy(int argc, VALUE argv[], VALUE self)
{
  RBFLTiled_Image *flimg = (RBFLTiled_Image*)rb_to_fl(self);
  RBFLImage *flimg2;
  VALUE w, h;

  switch( rb_scan_args(argc, argv, "02", &w, &h) ){
  case 0:
    flimg2 = (RBFLImage*)(flimg->copy());
    break;
  case 2:
    flimg2 = (RBFLImage*)(flimg->copy((int)NUM2INT(w), (int)NUM2INT(h)));
    break;
  default:
    rb_raise(rb_eArgError, "too few/many arguments");
  };

  return rb_FLImage_new(rb_cFLImage, flimg2);
};

static FN_CALL(Tiled_Image, desaturate);
static FN_SET_COLOR_DBL(Tiled_Image, color_average);


static VALUE rb_FLShared_Image_s_find(int argc, VALUE argv[], VALUE klass)
{
  VALUE img, w, h;
  Fl_Shared_Image *flimg;

  switch( rb_scan_args(argc, argv, "12", &img, &w, &h) ){
  case 1:
    flimg = RBFLShared_Image::find(STR2CSTR(img));
    break;
  case 3:
    flimg = RBFLShared_Image::find(STR2CSTR(img), NUM2INT(w), NUM2INT(h));
    break;
  default:
    rb_raise(rb_eArgError, "too few/many arguments");
  };

  return rb_FLImage_new(rb_cFLShared_Image, flimg);
};

static VALUE rb_FLShared_Image_s_get(int argc, VALUE argv[], VALUE klass)
{
  VALUE img, w, h;
  Fl_Shared_Image *flimg;

  switch( rb_scan_args(argc, argv, "12", &img, &w, &h) ){
  case 1:
    flimg = RBFLShared_Image::get(STR2CSTR(img));
    break;
  case 3:
    flimg = RBFLShared_Image::get(STR2CSTR(img), NUM2INT(w), NUM2INT(h));
    break;
  default:
    rb_raise(rb_eArgError, "too few/many arguments");
  };

  return rb_FLImage_new(rb_cFLShared_Image, flimg);
};

static VALUE rb_FLShared_Image_copy(int argc, VALUE argv[], VALUE self)
{
  RBFLShared_Image *flimg = (RBFLShared_Image*)rb_to_fl(self);
  RBFLImage *flimg2;
  VALUE w, h;

  switch( rb_scan_args(argc, argv, "02", &w, &h) ){
  case 0:
    flimg2 = (RBFLImage*)(flimg->copy());
    break;
  case 2:
    flimg2 = (RBFLImage*)(flimg->copy((int)NUM2INT(w), (int)NUM2INT(h)));
    break;
  default:
    rb_raise(rb_eArgError, "too few/many arguments");
  };

  return rb_FLImage_new(rb_cFLImage, flimg2);
};

static FN_GET_STR(Shared_Image, name);
static FN_GET_INT(Shared_Image, refcount);
static FN_CALL(Shared_Image, release);
static FN_CALL(Shared_Image, reload);
static FN_CALL(Shared_Image, desaturate);
static FN_SET_COLOR_DBL(Shared_Image, color_average);
#endif

extern "C" void
Init_flimage(){
  rb_cFLImage = rb_define_class_under(rb_mFLTK, "Image", rb_cObject);
  DEF_S_METHOD(Image, new, -1);
  DEF_INIT_METHOD(Image);
  DEF_METHOD(Image, destroy, 0);
  DEF_METHOD(Image, label, 1);
  DEF_METHOD(Image, draw, -1);

#if FLTK_VERSION >= FLTK_VERSION_CODE(1,1,0)
  rb_cFLRGB_Image = rb_define_class_under(rb_mFLTK, "RGBImage", rb_cFLImage);
  DEF_S_METHOD(RGB_Image, new, -1);
  DEF_METHOD(RGB_Image, destroy, 0);

  rb_cFLPNG_Image = rb_define_class_under(rb_mFLTK, "PNGImage", rb_cFLRGB_Image);
  DEF_S_METHOD(PNG_Image, new, -1);

  rb_cFLJPEG_Image = rb_define_class_under(rb_mFLTK, "JPEGImage", rb_cFLRGB_Image);
  DEF_S_METHOD(JPEG_Image, new, -1);

  rb_cFLBMP_Image = rb_define_class_under(rb_mFLTK, "BMPImage", rb_cFLRGB_Image);
  DEF_S_METHOD(BMP_Image, new, -1);

  rb_cFLTiled_Image = rb_define_class_under(rb_mFLTK, "TiledImage", rb_cFLImage);
  DEF_S_METHOD(Tiled_Image, new, -1);
  DEF_METHOD(Tiled_Image, copy, -1);
  DEF_METHOD(Tiled_Image, desaturate, 0);
  DEF_METHOD(Tiled_Image, color_average, 2);

  rb_cFLShared_Image = rb_define_class_under(rb_mFLTK, "SharedImage", rb_cFLImage);
  DEF_S_METHOD(Shared_Image, find, -1);
  DEF_S_METHOD(Shared_Image, get, -1);
  DEF_METHOD(Shared_Image, copy, -1);
  DEF_METHOD(Shared_Image, desaturate, 0);
  DEF_METHOD(Shared_Image, color_average, 2);
  DEF_METHOD(Shared_Image, name, 0);
  DEF_METHOD(Shared_Image, refcount, 0);
  DEF_METHOD(Shared_Image, release, 0);
  DEF_METHOD(Shared_Image, reload, 0);
#endif
};
