/* -*- C++ -*-
 * $Id: fltk.cc,v 1.159 2007/01/09 00:43:17 jeremy_henty Exp $
 */

#include "fltk.h"
#include "flenum.h"
#include "flcolor.h"
#include "flwidget.h"
#include "flbox.h"
#include "flbutton.h"
#include "flbrowser.h"
#include "flchart.h"
#include "flclock.h"
#include "flgroup.h"
#include "flcolorchooser.h"
#include "flpack.h"
#include "flscroll.h"
#include "fltabs.h"
#include "flspinner.h"
#include "fltile.h"
#include "flwindow.h"
#include "flinput.h"
#include "flmenu.h"
#include "flpositioner.h"
#include "flvaluator.h"
#include "flpixmap.h"
#include "flbitmap.h"
#include "flimage.h"

#include "fltext.h"
#include "flhelp.h"
#include "flfile.h"
#include "flwizard.h"
#include "fltooltip.h"
#include "flprogress.h"

#include <FL/x.H>
#include <FL/fl_file_chooser.H>
#include <FL/fl_ask.H>

#if defined(USE_GL)
# include <FL/gl.h>
# include <FL/glut.h>
#endif

/* test */
#if defined(USE_XSETLOCALE)
#  include <X11/Xlocale.h>
#elif defined(USE_SETLOCALE)
#  include <locale.h>
#endif


VALUE rb_mFLTK;
VALUE rb_mFLTKGL;
VALUE rb_eFLError;
static VALUE rb_fltk_objmap;
static VALUE rb_fltk_fdmap;
static VALUE rb_fltk_idle_callbacks;
static VALUE rb_fltk_timeout_callbacks;
static VALUE rb_fltk_check_callbacks;
static VALUE rb_fltk_handlers;
static double rb_fltk_thread_schedule_time;
static int rb_fltk_thread_schedule_freq = 0;

ID rb_id_handle;
ID rb_id_phandle;
ID rb_id_draw;
ID rb_id_pdraw;
ID rb_id_icallback;
ID rb_id_icallback_data;
ID rb_id_iitems;
ID rb_id_ilink;
ID rb_id_call;
ID rb_id_clear;
ID rb_id_delete;
ID rb_id_resize;
ID rb_id_flush;
ID rb_id_show;
ID rb_id_hide;
ID rb_id_draw_overlay;
static ID rb_id_index;
static ID rb_id_delete_if;

VALUE rb_ary_insert(VALUE ary, VALUE idx, VALUE data)
{
  rb_funcall(ary, rb_intern("[]="), 3, idx, INT2NUM(0), data);
  return Qnil;
};


#ifndef HAVE_RB_HASH_DELETE
static VALUE rb_hash_delete(VALUE hash, VALUE key)
{
  return rb_funcall(hash, rb_id_delete, 1, key);
}
#endif

#ifndef HAVE_RB_HASH_CLEAR
static VALUE rb_hash_clear(VALUE hash)
{
  return rb_funcall(hash, rb_id_clear, 0);
}
#endif /* HAVE_RB_HASH_CLEAR */

void rb_fltk_objmap_clear()
{
  rb_hash_clear(rb_fltk_objmap);
};

void rb_fltk_objmap_delete(Fl_Widget *w)
{
  rb_hash_delete(rb_fltk_objmap, LONG2NUM((long)w));
};

void rb_fltk_objmap_aset(Fl_Widget *w, VALUE obj)
{
  rb_hash_aset(rb_fltk_objmap, LONG2NUM((long)w), obj);
};

VALUE rb_fltk_objmap_aref(Fl_Widget *w)
{
  return rb_hash_aref(rb_fltk_objmap, LONG2NUM((long)w));
};

int *iary2cary(VALUE v, int last)
{
  int i, len;
  int *ary;
  VALUE e;

  if( v == Qnil ) return NULL;

  Check_Type(v, T_ARRAY);
  len = RARRAY(v)->len;
  ary = (int*)malloc(sizeof(long) * (len+1));
  for( i=0; i < len; i++ ){
    e = rb_ary_entry(v, i);
    switch( TYPE(e) ){
    case T_FIXNUM:
    case T_BIGNUM:
      ary[i] = (int)(NUM2INT(e));
      break;
    case T_NIL:
      ary[i] = 0;
      break;
    default:
      rb_raise(rb_eTypeError, "unexpected type of the element #%d", i);
      break;
    };
  };
  ary[len] = last;

  return ary;
};

VALUE cary2iary(int *is, int last)
{
  VALUE ary = rb_ary_new();

  if( !is ) return Qnil;

  for( int i=0; is[i] != last; i++ ){
    rb_ary_push(ary, INT2NUM(is[i]));
  };

  return ary;
};

VALUE cary2iary_n(int *is, int len)
{
  VALUE ary = rb_ary_new();

  if( !is ) return Qnil;

  for( int i=0; i < len; i++ ){
    rb_ary_push(ary, INT2NUM(is[i]));
  };

  return ary;
};

float *fary2cary(VALUE v, float last)
{
  int i, len;
  float *ary;
  VALUE e;

  if( v == Qnil ) return NULL;

  Check_Type(v, T_ARRAY);
  len = RARRAY(v)->len + 1;
  ary = (float*)malloc(sizeof(double) * len);
  for( i=0; i < len - 1; i++ ){
    e = rb_ary_entry(v, i);
    switch( TYPE(e) ){
    case T_FLOAT:
      ary[i] = (float)(RFLOAT(e)->value);
      break;
    case T_NIL:
      ary[i] = 0.0;
      break;
    default:
      rb_raise(rb_eTypeError, "unexpected type of the element #%d", i);
      break;
    };
  };
  ary[len] = last;

  return ary;
};

VALUE cary2fary_n(float *fs, int len)
{
  VALUE ary = rb_ary_new();

  if( !fs ) return Qnil;

  for( int i=0; i < len; i++ ){
    rb_ary_push(ary, rb_float_new(fs[i]));
  };

  return ary;
};

char ** rb_fltk_create_pixmap(VALUE ary)
{
  int len;
  char ** data;

  Check_Type(ary, T_ARRAY);
  len = RARRAY(ary)->len;
  if( len <= 0 ) return NULL;

  data = (char**)malloc(sizeof(char*) * (len + 1));
  for( int i=0; i<=len-1; i++ ){
    VALUE str = RARRAY(ary)->ptr[i];
    Check_Type(str, T_STRING);
    data[i] = (char*)malloc(RSTRING(str)->len+1);
    memcpy(data[i], STR2CSTR(str), RSTRING(str)->len);
    data[i][RSTRING(str)->len] = '\0';
  };
  data[len] = NULL;

  return (char**)data;
};

char *rb_fltk_create_bitmap(VALUE ary)
{
  int len;
  char *data;

  Check_Type(ary, T_ARRAY);
  len = RARRAY(ary)->len;
  data = (char*)malloc(len);
  for( int i=0; i<=len-1; i++ ){
    VALUE c = RARRAY(ary)->ptr[i];
    data[i] = (char)NUM2INT(c);
  };

  return data;
};

uchar *rb_fltk_create_image(VALUE ary)
{
  int len;
  uchar *data;

  Check_Type(ary, T_ARRAY);
  len = RARRAY(ary)->len;
  data = (uchar*)malloc(len);
  for( int i=0; i<=len-1; i++ ){
    VALUE c = RARRAY(ary)->ptr[i];
    data[i] = NUM2UCHAR(c);
  };

  return data;
};


static void rb_fltk_call_thread_schedule(void *data)
{
  if( rb_fltk_thread_schedule_time > 0.0 ){
    Fl::add_timeout(rb_fltk_thread_schedule_time, rb_fltk_call_thread_schedule);
    if( rb_fltk_thread_schedule_freq > 0 ){
      for( int i=0; i<rb_fltk_thread_schedule_freq; i++ )
	rb_thread_schedule();
    }
    else{
      rb_thread_schedule();
    };
  };
};

static void rb_fltk_call_thread_schedule_once(void *data)
{
  rb_thread_schedule();
};

static VALUE rb_fltk_enable_thread(int argc, VALUE argv[], VALUE self)
{
  VALUE val;
  int add = 0;

  if( rb_fltk_thread_schedule_time <= 0.0 ) add = 1;

  switch( rb_scan_args(argc, argv, "01", &val) ){
  case 0:
    rb_fltk_thread_schedule_time = RBFL_THREAD_SCHEDULE_TIME;
    break;
  case 1:
    Check_Type(val, T_FLOAT);
    rb_fltk_thread_schedule_time = RFLOAT(val)->value;
    break;
  };

  if( add ){
    Fl::add_timeout(rb_fltk_thread_schedule_time, rb_fltk_call_thread_schedule);
  };

  return Qnil;
};

static VALUE rb_fltk_disable_thread(VALUE self)
{
  rb_fltk_thread_schedule_time = -1.0;
  return Qnil;
};

static VALUE rb_fltk_thread_freq(int argc, VALUE argv[], VALUE mod)
{
  VALUE freq;

  switch( rb_scan_args(argc, argv, "01", &freq) ){
  case 0:
    return INT2NUM(rb_fltk_thread_schedule_freq);
  case 1:
    rb_fltk_thread_schedule_freq = NUM2INT(freq);
    return freq;
  };
};

static void rb_fltk_add_fd_func(int fd, void *data)
{
  VALUE proc;

  proc = rb_hash_aref(rb_fltk_fdmap, INT2NUM(fd));
  rb_funcall(proc, rb_id_call, 2, INT2NUM(fd), CPTR2VAL(data));
};

static VALUE rb_fltk_add_fd(int argc, VALUE argv[], VALUE mod)
{
  VALUE arg1, arg2, arg3, arg4;
  int fd = -1;
  int when;
  VALUE data, proc;

  if( rb_block_given_p() ){
    switch( rb_scan_args(argc, argv, "12", &arg1, &arg2, &arg3) ){
    case 1:
      fd = NUM2INT(rb_Integer(arg1));
      proc = rb_block_proc();
      data = Qnil;
      Fl::add_fd(fd, rb_fltk_add_fd_func, VAL2CPTR(data));
      break;
    case 2:
      fd = NUM2INT(rb_Integer(arg1));
      proc = rb_block_proc();
      data = arg2;
      Fl::add_fd(fd, rb_fltk_add_fd_func, VAL2CPTR(data));
      break;
    case 3:
      fd = NUM2INT(rb_Integer(arg1));
      when = NUM2INT(arg2);
      proc = rb_block_proc();
      data = arg3;
      Fl::add_fd(fd, when, rb_fltk_add_fd_func, VAL2CPTR(data));
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "22", &arg1, &arg2, &arg3, &arg4) ){
    case 2:
      fd = NUM2INT(rb_Integer(arg1));
      proc = arg2;
      data = Qnil;
      Fl::add_fd(fd, rb_fltk_add_fd_func, VAL2CPTR(data));
      break;
    case 3:
      fd = NUM2INT(rb_Integer(arg1));
      proc = arg2;
      data = arg3;
      Fl::add_fd(fd, rb_fltk_add_fd_func, VAL2CPTR(data));
      break;
    case 4:
      fd = NUM2INT(rb_Integer(arg1));
      when = NUM2INT(arg2);
      proc = arg3;
      data = arg4;
      Fl::add_fd(fd, when, rb_fltk_add_fd_func, VAL2CPTR(data));
      break;
    };
  };

  rb_hash_aset(rb_fltk_fdmap, INT2NUM(fd), proc);

  return rb_ary_new3(3, INT2NUM(fd), proc, data);
};

static VALUE rb_fltk_remove_fd(VALUE mod, VALUE fd)
{
  Fl::remove_fd(NUM2INT(fd));
  rb_hash_delete(rb_fltk_fdmap, fd);
  return Qnil;
};

static void rb_fltk_idle_callback(void *user_data)
{
  VALUE proc = rb_ary_entry(CPTR2VAL(user_data), 0);
  VALUE data = rb_ary_entry(CPTR2VAL(user_data), 1);

  if( rb_ary_includes(rb_fltk_idle_callbacks, CPTR2VAL(user_data)) == Qtrue ){
    rb_apply(proc, rb_id_call, data);
  };
};

static void rb_fltk_timeout_callback(void *user_data)
{
  VALUE proc = rb_ary_entry(CPTR2VAL(user_data), 0);
  VALUE data = rb_ary_entry(CPTR2VAL(user_data), 1);
  VALUE idx;

  idx = rb_funcall(rb_fltk_timeout_callbacks, rb_id_index, 1, CPTR2VAL(user_data));
  if( idx != Qnil ){
    rb_apply(proc, rb_id_call, data);
    rb_ary_delete_at(rb_fltk_timeout_callbacks, NUM2INT(idx));
  };
};

static void rb_fltk_check_callback(void *user_data)
{
  VALUE proc = rb_ary_entry(CPTR2VAL(user_data), 0);
  VALUE data = rb_ary_entry(CPTR2VAL(user_data), 1);

  if( rb_ary_includes(rb_fltk_check_callbacks, CPTR2VAL(user_data)) == Qtrue ){
    rb_apply(proc, rb_id_call, data);
  };
};

static VALUE rb_fltk_add_idle(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc, data, user_data;

  if( rb_block_given_p() ){
    rb_scan_args(argc, argv, "*&", &data, &proc);
  } else {
    rb_scan_args(argc, argv, "1*", &proc, &data);
  }

  user_data = rb_assoc_new(proc, data);
  rb_ary_push(rb_fltk_idle_callbacks, user_data);
  Fl::add_idle(rb_fltk_idle_callback, VAL2CPTR(user_data));
  return user_data;
};

static VALUE rb_fltk_remove_idle(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc, data, user_data, idx;

  rb_scan_args(argc, argv, "1*", &proc, &data);

  user_data = rb_assoc_new(proc, data);
  idx = rb_funcall(rb_fltk_idle_callbacks, rb_id_index, 1, user_data);
  if( idx != Qnil ){
    user_data = rb_ary_entry(rb_fltk_idle_callbacks, NUM2INT(idx));
    Fl::remove_idle(rb_fltk_idle_callback, VAL2CPTR(user_data));
    rb_ary_delete_at(rb_fltk_idle_callbacks, NUM2INT(idx));
  };
  return Qnil;
};

static VALUE rb_fltk_add_timeout(int argc, VALUE argv[], VALUE mod)
{
  VALUE time, proc, data, user_data;

  if( rb_block_given_p() ){
    rb_scan_args(argc, argv, "1*&", &time, &data, &proc);
  } else {
    rb_scan_args(argc, argv, "2*", &time, &proc, &data);
  }

  time = rb_Float(time);

  user_data = rb_assoc_new(proc, data);
  rb_ary_push(rb_fltk_timeout_callbacks, user_data);
  Fl::add_timeout(RFLOAT(time)->value, rb_fltk_timeout_callback, VAL2CPTR(user_data));
  return user_data;
};

static VALUE rb_fltk_repeat_timeout(int argc, VALUE argv[], VALUE mod)
{
  VALUE time, proc, data, user_data;

  if( rb_block_given_p() ){
    rb_scan_args(argc, argv, "1*&", &time, &data, &proc);
  } else {
    rb_scan_args(argc, argv, "2*", &time, &proc, &data);
  }

  time = rb_Float(time);

  user_data = rb_assoc_new(proc, data);
  rb_ary_push(rb_fltk_timeout_callbacks, user_data);
  Fl::repeat_timeout(RFLOAT(time)->value, rb_fltk_timeout_callback, VAL2CPTR(user_data));
  return user_data;
};

static VALUE delete_if(VALUE ary)
{
  return rb_funcall(ary,rb_id_delete_if,0);
}

static VALUE delete_timeout(VALUE data, VALUE user_data){
  VALUE is_equal = rb_equal(data,user_data);
  if ( is_equal ) {
    Fl::remove_timeout(rb_fltk_timeout_callback, VAL2CPTR(data));
  }
  return is_equal;
}

static VALUE rb_fltk_remove_timeout(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc, data, user_data, idx;

  rb_scan_args(argc, argv, "1*", &proc, &data);

  user_data = rb_assoc_new(proc, data);
  rb_iterate(delete_if,rb_fltk_timeout_callbacks,
	     (VALUE(*)(ANYARGS)) delete_timeout,user_data);
  return Qnil;
};

static VALUE rb_fltk_add_check(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc, data, user_data;

  if( rb_block_given_p() ){
    rb_scan_args(argc, argv, "*&", &data, &proc);
  } else {
    rb_scan_args(argc, argv, "1*", &proc, &data);
  }

  user_data = rb_assoc_new(proc, data);
  rb_ary_push(rb_fltk_check_callbacks, user_data);
  Fl::add_check(rb_fltk_check_callback, VAL2CPTR(user_data));
  return user_data;
};

static VALUE delete_check(VALUE data, VALUE user_data){
  VALUE is_equal = rb_equal(data,user_data);
  if ( is_equal ) {
    Fl::remove_check(rb_fltk_check_callback, VAL2CPTR(data));
  }
  return is_equal;
}

static VALUE rb_fltk_remove_check(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc, data, user_data, idx;

  rb_scan_args(argc, argv, "1*", &proc, &data);

  user_data = rb_assoc_new(proc, data);
  rb_iterate(delete_if,rb_fltk_check_callbacks,
	     (VALUE(*)(ANYARGS)) delete_check,user_data);
  return Qnil;
};

static int rb_fltk_internal_handler(int e)
{
  int   len = RARRAY(rb_fltk_handlers)->len;
  VALUE *ptr    = RARRAY(rb_fltk_handlers)->ptr;
  VALUE proc;
  VALUE r;

  for( int i = 0; i < len; i++ ){
    proc = ptr[i];
    r = rb_funcall(proc, rb_id_call, 1, INT2NUM(e));
    rb_thread_schedule();
    if( r == Qtrue ){
      return 1;
    };
  };
  return 0;
};

static VALUE rb_fltk_add_handler(int argc, VALUE argv[], VALUE mod)
{
  VALUE proc;

  if( rb_block_given_p() ){
    if( argc == 0 ){
      proc = rb_block_proc();
    }
    else{
      rb_raise(rb_eArgError, "too many arguments");
    };
  }
  else{
    if( argc == 1 ){
      proc = argv[0];
    }
    else{
      rb_raise(rb_eArgError, "too few/many arguments");
    };
  };

  rb_ary_push(rb_fltk_handlers, proc);

  return Qnil;
};

static VALUE rb_fltk_draw(int argc, VALUE argv[], VALUE self)
{
  VALUE str, x, y, n, w, h, align;

  if( fl_size() <= 0 ){
    rb_warn("The size is zero. make it non-zero before drawing.\n");
    return Qnil;
  };
  switch( rb_scan_args(argc, argv, "33", &str, &x, &y, &w, &h, &align) ){
  case 3:
    fl_draw((const char *)STR2CSTR(str), NUM2INT(x), NUM2INT(y));
    break;
  case 4:
    n = x; x = y; y = w;
    fl_draw((const char *)STR2CSTR(str), NUM2INT(n),
	    NUM2INT(x), NUM2INT(y));
    break;
  case 5:
    fl_draw((const char *)STR2CSTR(str), NUM2INT(x), NUM2INT(y),
	    NUM2INT(w), NUM2INT(h), FL_ALIGN_CENTER);
    break;
  case 6:
    fl_draw((const char *)STR2CSTR(str), NUM2INT(x), NUM2INT(y),
	    NUM2INT(w), NUM2INT(h), (Fl_Align)NUM2INT(align));
    break;
  };
  return Qnil;
};

static VALUE rb_fltk_width(int argc, VALUE argv[], VALUE self)
{
  VALUE str, n;
  double ret;

  switch( rb_scan_args(argc, argv, "11", &str, &n) ){
  case 1:
    ret = fl_width(STR2CSTR(str));
    break;
  case 2:
    ret = fl_width(STR2CSTR(str), NUM2INT(n));
    break;
  };
  return rb_float_new(ret);
};

static FN_GLOBAL_GET_INT(height);
static FN_GLOBAL_GET_INT(descent);

static VALUE rb_fltk_color(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1, arg2, arg3;

  switch( rb_scan_args(argc, argv, "03", &arg1, &arg2, &arg3) ){
  case 0:
    return COLOR2VALUE(fl_color());
    break; 
  case 1:
    fl_color(VALUE2COLOR(arg1));
    break;
  case 2:
    rb_raise(rb_eArgError, "wrong number of arguments (2 for 0, 1 or 3)");
    break;
  case 3:
    fl_color(NUM2UCHAR(arg1),NUM2UCHAR(arg2),NUM2UCHAR(arg3));
    break;
  };
  return Qnil;
};

static VALUE rb_fltk_clip_box(VALUE mod, VALUE x, VALUE y, VALUE w, VALUE h)
{
  int rx,ry,rw,rh;

  fl_clip_box(NUM2INT(x), NUM2INT(y), NUM2INT(w), NUM2INT(h),
	      rx, ry, rw, rh);
  return rb_ary_new3(4, INT2NUM(rx), INT2NUM(ry), INT2NUM(rw), INT2NUM(rh));
};

static char *obj_type_name(VALUE x) {
  // copied from rb_check_type in error.c 
  char *type;
  if (NIL_P(x)) {
    type = "nil";
  }
  else if (FIXNUM_P(x)) {
    type = "Fixnum";
  }
  else if (SYMBOL_P(x)) {
    type = "Symbol";
  }
  else if (rb_special_const_p(x)) {
    type = RSTRING(rb_obj_as_string(x))->ptr;
  }
  else {
    type = rb_obj_classname(x);
  }
  return type;
}

static VALUE rb_fltk_line_style(VALUE argc, VALUE argv[], VALUE self)
{
    VALUE arg1, arg2, arg3;
    char *dashes, *bad_type;
    int len;

    switch( rb_scan_args(argc, argv, "12", &arg1, &arg2, &arg3) ){
      case 1:
        fl_line_style(NUM2INT(arg1));
	break;
      case 2:
        fl_line_style(NUM2INT(arg1), NUM2INT(arg2));
	break;
      case 3:
	switch( TYPE(arg3) ){
	case T_NIL:
	  dashes = 0;
	  break;
	case T_STRING:
	  dashes = STR2CSTR(arg3);
	  break;
	case T_ARRAY:
	  len = RARRAY(arg3)->len;
	  dashes = ALLOCA_N(char, len + 1);
	  for( int i = 0; i < len; i++ ){
	    VALUE w = RARRAY(arg3)->ptr[i];
	    switch( TYPE(w) ){
	    case T_FIXNUM:
	      dashes[i] = NUM2INT(w);
	      break;
	    case T_STRING:
	      dashes[i] = char(STR2CSTR(w)[0]);
	      break;
	    default:
	      rb_raise(rb_eTypeError,
		       "wrong argument type %s (expected Fixnum or String)",
		       obj_type_name(w)); 
	      break;
	    };
	  };
	  dashes[len] = 0;
	  break;
	default:
	  rb_raise(rb_eTypeError,
		   "wrong argument type %s (expected nil, Array or String)",
		   obj_type_name(arg3));
	  break;
	};
	fl_line_style(NUM2INT(arg1), NUM2INT(arg2), dashes);
        break;
    };
    return Qnil;
};

static VALUE rb_fltk_set_font(VALUE self, VALUE arg1, VALUE arg2)
{
  Check_Type(arg2,T_STRING);
  Fl::set_font((Fl_Font)NUM2INT(arg1), STR2CSTR(arg2));

  return Qnil;
};

static VALUE rb_fltk_set_fonts(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1;
  int ret;
  switch( rb_scan_args(argc, argv, "01", &arg1) ){
  case 0:
    ret = Fl::set_fonts();
    break;
  case 1:
    ret = Fl::set_fonts(STR2CPTR(arg1));
    break;
  };
  return INT2NUM(ret);
};

static VALUE rb_fltk_event_key(int argc, VALUE argv[], VALUE self)
{
  VALUE ch;

  switch( rb_scan_args(argc, argv, "01", &ch) ){
  case 0:
    return INT2NUM(Fl::event_key());
  case 1:
    return INT2BOOL(Fl::event_key((int)NUM2INT(ch)));
  };
};

static VALUE rb_fltk_event_clicks(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1;

  switch( rb_scan_args(argc, argv, "01", &arg1) ){
  case 0:
    return INT2NUM(Fl::event_clicks());
  case 1:
    Fl::event_clicks(NUM2INT(arg1));
    return Qnil;
  };
};

static VALUE rb_fltk_event_inside(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1, arg2, arg3, arg4;
  Fl_Widget *w;

  switch( rb_scan_args(argc, argv, "13", &arg1, &arg2, &arg3, &arg4) )
  {
  case 1:
    rb_fl_check_widget(arg1);
    w = rb_to_fl(arg1);
    return INT2BOOL(Fl::event_inside(w));
  case 4:
    return INT2BOOL(Fl::event_inside(NUM2INT(arg1), NUM2INT(arg2),
	NUM2INT(arg3), NUM2INT(arg4)));
  default:
    rb_raise(rb_eArgError, "wrong number of argumments (must be 1 or 4)");
  };
};

static VALUE rb_fltk_event_is_click(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1;

  switch( rb_scan_args(argc, argv, "01", &arg1) )
  {
  case 0:
    return INT2BOOL(Fl::event_is_click());
  case 1:
    Fl::event_is_click(0);
    return Qnil;
  };
};

static VALUE rb_fltk_event_state(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1;

  switch( rb_scan_args(argc, argv, "01", &arg1) )
  {
  case 0:
    return INT2NUM(Fl::event_state());
  case 1:
    return INT2BOOL(Fl::event_state(NUM2INT(arg1)));
  };
};

static VALUE rb_fltk_size(int argc, VALUE argv[], VALUE self)
{
  VALUE size;

  switch( rb_scan_args(argc, argv, "01", &size) ){
  case 0:
    size = INT2NUM(fl_size());
    return size;
  case 1:
    fl_font(fl_font(), NUM2INT(size));
    return Qnil;
  };
};

static VALUE rb_fltk_font(int argc, VALUE argv[], VALUE self)
{
  VALUE font, size;

  switch( rb_scan_args(argc, argv, "02", &font, &size) ){
  case 0:
    font = INT2NUM(fl_font());
    return font;
  case 1:
    fl_font(NUM2INT(font), fl_size());
    return Qnil;
  case 2:
    fl_font(NUM2INT(font), NUM2INT(size));
    return Qnil;
  };
};

static VALUE rb_fltk_run(VALUE self)
{
  Fl::run();
  rb_fltk_objmap_clear();
  return Qnil;
};

static FN_MODULE_FL_CALL(check);
static FN_MODULE_FL_CALL(flush);
static FN_MODULE_FL_GETSET_WIDGET(focus);
static FN_MODULE_FL_GET_INT(event);
static FN_MODULE_FL_GET_INT(event_x);
static FN_MODULE_FL_GET_INT(event_y);
static FN_MODULE_FL_GET_INT(event_x_root);
static FN_MODULE_FL_GET_INT(event_y_root);
#ifdef HAVE_FL_EVENT_DX
static FN_MODULE_FL_GET_INT(event_dx);
static FN_MODULE_FL_GET_INT(event_dy);
#endif

static FN_MODULE_FL_GET_INT(event_length);
static FN_MODULE_FL_GET_STR(event_text,event_length);
static FN_MODULE_FL_GET_BOOL(event_shift);
static FN_MODULE_FL_GET_BOOL(event_ctrl);
static FN_MODULE_FL_GET_BOOL(event_alt);
static FN_MODULE_FL_GET_INT(event_button);
static FN_MODULE_FL_GET_BOOL(event_button1);
static FN_MODULE_FL_GET_BOOL(event_button2);
static FN_MODULE_FL_GET_BOOL(event_button3);
static FN_MODULE_FL_GET_INT_SET_INT(get_key);
static FN_MODULE_FL_SET_UCHAR3(foreground);
static FN_MODULE_FL_SET_UCHAR3(background);
static FN_MODULE_FL_SET_UCHAR3(background2);
static FN_MODULE_FL_GET_BOOL_SET_INT(visual);
static FN_MODULE_FL_GET_INT(x);
static FN_MODULE_FL_GET_INT(y);
static FN_MODULE_FL_GET_INT(w);
static FN_MODULE_FL_GET_INT(h);
static FN_GLOBAL_SET_INT4_OPTINT2(line);

static VALUE rb_fltk_xyline(int argc, VALUE argv[], VALUE self) {
  if (argc <= 2)
    rb_raise(rb_eArgError, "too few arguments, %d for 3 or more", argc);
  while (1) {
    switch (argc) {
    case 3:
      fl_xyline(NUM2INT(argv[0]), NUM2INT(argv[1]), 
		NUM2INT(argv[2]));
      return Qnil;
    case 4:
      fl_xyline(NUM2INT(argv[0]), NUM2INT(argv[1]),
                NUM2INT(argv[2]), NUM2INT(argv[3]));
      return Qnil; 
    case 5:
      fl_xyline(NUM2INT(argv[0]), NUM2INT(argv[1]),
                NUM2INT(argv[2]), NUM2INT(argv[3]), NUM2INT(argv[4]));
      return Qnil; 
    default:
      fl_xyline(NUM2INT(argv[0]), NUM2INT(argv[1]),
                NUM2INT(argv[2]), NUM2INT(argv[3]));
      argc -= 2;
      argv += 2;
    }
  }
  return Qnil; // never reached 
}

static VALUE rb_fltk_yxline(int argc, VALUE argv[], VALUE self) {
  if (argc <= 2)
    rb_raise(rb_eArgError, "too few arguments, %d for 3 or more", argc);
  while (1) {
    switch (argc) {
    case 3:
      fl_yxline(NUM2INT(argv[0]), NUM2INT(argv[1]), 
		NUM2INT(argv[2]));
      return Qnil;
    case 4:
      fl_yxline(NUM2INT(argv[0]), NUM2INT(argv[1]),
                NUM2INT(argv[2]), NUM2INT(argv[3]));
      return Qnil; 
    case 5:
      fl_yxline(NUM2INT(argv[0]), NUM2INT(argv[1]),
                NUM2INT(argv[2]), NUM2INT(argv[3]), NUM2INT(argv[4]));
      return Qnil; 
    default:
      fl_yxline(NUM2INT(argv[0]), NUM2INT(argv[1]),
                NUM2INT(argv[2]), NUM2INT(argv[3]));
      argc -= 2;
      argv += 2;
      // swap the first two remaining arguments 
      VALUE tmp = argv[0];
      argv[0] = argv[1];
      argv[1] = tmp;
    }
  }
  return Qnil; // never reached 
}

static FN_GLOBAL_SET_INT4(push_clip);
static FN_GLOBAL_CALL(push_no_clip);
static FN_GLOBAL_CALL(pop_clip);
static FN_GLOBAL_GET_BOOL_SET_INT4(not_clipped);
static FN_GLOBAL_SET_INT4(rect);
static FN_GLOBAL_SET_INT4_OPTCHR3(rectf);
static FN_GLOBAL_SET_INT4_DBL2(pie);
static FN_GLOBAL_SET_INT4_DBL2(arc);
static FN_GLOBAL_SET_DBL3(circle);
static FN_GLOBAL_SET_DBL2(vertex);
static FN_GLOBAL_CALL(begin_line);
static FN_GLOBAL_CALL(end_line);
static FN_GLOBAL_CALL(begin_loop);
static FN_GLOBAL_CALL(end_loop);
static FN_GLOBAL_CALL(begin_polygon);
static FN_GLOBAL_CALL(end_polygon);
static FN_GLOBAL_CALL(begin_complex_polygon);
static FN_GLOBAL_CALL(end_complex_polygon);
static FN_GLOBAL_CALL(gap);
static FN_GLOBAL_SET_INT2(point);
static FN_GLOBAL_SET_FORMAT(message);
static FN_GLOBAL_SET_FORMAT(alert);
static FN_GLOBAL_GET_BOOL_SET_FORMAT(ask);
static FN_GLOBAL_GET_INT_SET_FORMAT_STR3(choice);
static FN_GLOBAL_GET_STR_SET_FORMAT_STR(input);
static FN_GLOBAL_GET_STR_SET_FORMAT_STR(password);
static FN_GLOBAL_GET_COLOR_SET_COLOR(lighter);
static FN_GLOBAL_GET_COLOR_SET_COLOR(darker);

static VALUE rb_fltk_beep(int argc, VALUE argv[], VALUE self)
{
  VALUE ty;

  switch ( rb_scan_args(argc, argv, "01", &ty) ) {
  case 0:
    fl_beep();
    break;
  case 1:
    fl_beep(NUM2INT(ty));
    break;
  }
  return Qnil; 
}

static VALUE rb_fltk_rgb_color(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1, arg2, arg3;
  Fl_Color r;

  switch(rb_scan_args(argc, argv, "12", &arg1, &arg2, &arg3))
  {
    case 1:
	r = fl_rgb_color(NUM2UCHAR(arg1));
	break;
    case 2:
	rb_raise(rb_eArgError, "wrong number of arguments (2 for 3)");    
    case 3:
	r = fl_rgb_color(NUM2UCHAR(arg1), NUM2UCHAR(arg2), NUM2UCHAR(arg3));
  };
  return COLOR2VALUE(r);
};



static VALUE rb_fltk_wait(int argc, VALUE argv[], VALUE self)
{
VALUE time;
  switch(rb_scan_args(argc,argv,"01",&time))
  {
    case 0:
      return INT2BOOL(Fl::wait());
    case 1:
      time = rb_Float(time);
      return rb_float_new(Fl::wait(RFLOAT(time)->value));
  }
  return Qnil;
}

static VALUE rb_fltk_draw_pixmap(int argc, VALUE argv[], VALUE self)
{
  VALUE xpm, x, y, color;
  int r;
  char **data;

  rb_scan_args(argc, argv, "31", &xpm, &x, &y, &color);
  data = rb_fltk_create_pixmap(xpm);
  switch( argc ){
  case 3:
    r = INT2NUM(fl_draw_pixmap(data, NUM2INT(x), NUM2INT(y)));
    break;
  case 4:
    r = INT2NUM(fl_draw_pixmap(data, NUM2INT(x), NUM2INT(y),
			       VALUE2COLOR(color)));
    break;
  };
  return r;
};

static VALUE rb_fltk_measure_pixmap(VALUE self, VALUE xpm)
{
  VALUE str;
  char *data[2];
  int r, w, h;

  Check_Type(xpm, T_ARRAY);
  str = RARRAY(xpm)->ptr[0];
  data[0] = STR2CSTR(str);
  data[1] = NULL;
  r = fl_measure_pixmap(data, w, h);
  if( r ){
    return rb_ary_new3(2, INT2NUM(w), INT2NUM(h));
  }
  else{
    return Qnil;
  };
};

#ifdef HAVE_FL_READ_IMAGE
static VALUE rb_fltk_read_image(int argc, VALUE argv[], VALUE mod)
{
  VALUE vx, vy, vw, vh, valpha;
  int   x, y, w, h, alpha;
  uchar *p;
  VALUE v;

  x = y = w = h = alpha = 0;
  switch( rb_scan_args(argc, argv, "23", &vx, &vy, &vw, &vh, &valpha) ){
  case 5:
    alpha = NUM2INT(valpha);
  case 4:
    h = NUM2INT(vh);
  case 3:
    w = NUM2INT(vw);
  case 2:
    y = NUM2INT(vy);
    x = NUM2INT(vx);
  };
  p = fl_read_image(NULL, x, y, w, h, alpha);
  if( p ){
    v = rb_tainted_str_new2((const char *)p);
    delete p;
  }
  else{
    v = Qnil;
  }
  return v;
}
#endif

static VALUE rb_fltk_file_chooser_callback_obj = Qnil;

static void rb_fltk_file_chooser_callback_internal(const char *str)
{
  if( rb_fltk_file_chooser_callback_obj != Qnil ){
    rb_funcall(rb_fltk_file_chooser_callback_obj, rb_id_call, 1, CPTR2STR(str));
  };
};

static VALUE rb_fltk_file_chooser_callback(int argc, VALUE argv[], VALUE self)
{
  VALUE proc;

  switch( rb_scan_args(argc, argv, "01", proc) ){
  case 0:
    proc = rb_block_proc();
    break;
  case 1:
    /* do nothing */
    break;
  };
  rb_fltk_file_chooser_callback_obj = proc;
  fl_file_chooser_callback(rb_fltk_file_chooser_callback_internal);

  return Qnil;
};

static VALUE rb_fltk_file_chooser(int argc, VALUE argv[], VALUE self)
{
  VALUE msg, pat, fname;
  const char *c_msg, *c_pat, *c_fname;
  char *ret;

  switch( rb_scan_args(argc, argv, "12", &msg, &pat, &fname) ){
  case 1:
    c_msg = STR2CPTR(msg);
    c_pat = NULL;
    c_fname = NULL;
    break;
  case 2:
    c_msg = STR2CPTR(msg);
    c_pat = STR2CPTR(pat);
    c_fname = NULL;
    break;
  case 3:
    c_msg = STR2CPTR(msg);
    c_pat = STR2CPTR(pat);
    c_fname = STR2CPTR(fname);
    break;
  };

  ret = fl_file_chooser(c_msg, c_pat, c_fname);
  return ret ? rb_tainted_str_new2(ret) : Qnil;
};

static VALUE rb_fltk_dir_chooser(int argc, VALUE argv[], VALUE self)
{
  VALUE msg, fname;
  const char *c_msg, *c_fname;
  char *ret;

  switch( rb_scan_args(argc, argv, "11", &msg, &fname) ){
  case 1:
    c_msg = STR2CPTR(msg);
    c_fname = NULL;
    break;
  case 2:
    c_msg = STR2CPTR(msg);
    c_fname = STR2CPTR(fname);
    break;
  };

  ret = fl_dir_chooser(c_msg, c_fname);
  return ret ? rb_tainted_str_new2(ret) : Qnil;
};

static VALUE rb_fltk_paste( VALUE self, VALUE arg1 ){
  Fl_Widget *widget;

  widget =  rb_to_fl(arg1);
  if( widget ){
    if(Fl::check()){
      Fl::paste(*widget);
    }
    else{
      rb_raise(rb_eFLError, "FLTK does not appear to be taking events!");
    };
  }
  else{
    rb_raise(rb_eTypeError, "Need widget to send paste-event to.");
  };
  return Qnil;
};

static VALUE rb_fltk_xid(VALUE mod, VALUE win)
{
  Fl_Window *flwin;

  if( !rb_obj_is_kind_of(win, rb_cFLWindow) ){
    rb_raise(rb_eTypeError, "must be FLTK::Window object.");
  };
  flwin = (Fl_Window*)rb_to_fl(win);
#if defined(WIN32)
  return INT2NUM((unsigned long)fl_xid(flwin));
#else
  return INT2NUM(fl_xid(flwin));
#endif
};

static VALUE rb_fltk_find(VALUE mod, VALUE xid)
{
  Fl_Window *flwin;

  flwin = fl_find((Window)NUM2INT(xid));
  return rb_fltk_objmap_aref(flwin);
};

#if !defined(WIN32)
static VALUE rb_fltk_set_xid(VALUE mod, VALUE win, VALUE xid)
{
  Fl_Window *flwin;

  if( !rb_obj_is_kind_of(win, rb_cFLWindow) ){
    rb_raise(rb_eTypeError, "must be FLTK::Window object.");
  };
  flwin = (Fl_Window*)rb_to_fl(win);
  if( Fl_X::set_xid(flwin, (Window)INT2NUM(xid)) ){
    return Qtrue;
  }
  else{
    return Qfalse;
  };
};
#endif

#ifdef USE_UTF8
#include <FL/fl_utf8.h>
static VALUE rb_fltk_utflen(VALUE self, VALUE str)
{
  Check_Type(str, T_STRING);
  return INT2NUM(fl_utflen((const unsigned char *)STR2CSTR(str),
			   RSTRING(str)->len));
};
static VALUE rb_fltk_utf_nb_char(VALUE self, VALUE str)
{
  Check_Type(str, T_STRING);
  return INT2NUM(fl_utflen((const unsigned char *)STR2CSTR(str),
			   RSTRING(str)->len));
};
#endif

#ifdef USE_GL
static FN_GL_GLOBAL_SET_COLOR(color);
#endif

static VALUE rb_fltk_get_color(VALUE self, VALUE arg1)
{
  return UINT2NUM(Fl::get_color(VALUE2COLOR(arg1)));
};

static VALUE rb_fltk_get_color_rgb(VALUE self, VALUE arg1)
{
  uchar r, g, b;
  Fl::get_color(VALUE2COLOR(arg1),r,g,b);
  return rb_ary_new3(3,UCHAR2NUM(r),UCHAR2NUM(g),UCHAR2NUM(b)); 
};

static VALUE rb_fltk_set_color(int argc, VALUE argv[], VALUE self)
{
  VALUE arg1, arg2, arg3, arg4;

  switch(rb_scan_args(argc, argv, "22", &arg1, &arg2, &arg3, &arg4))
  {
    case 2:
      Fl::set_color(VALUE2COLOR(arg1), NUM2UINT(arg2));
      break;
    case 4:
      Fl::set_color(VALUE2COLOR(arg1),
        NUM2UCHAR(arg2), NUM2UCHAR(arg3), NUM2UCHAR(arg4));
      break;
    default:
      rb_raise(rb_eArgError, "wrong number of arguments (must be 2 or 4)");
  };

  return Qnil;
};

static VALUE rb_fltk_color_average(VALUE self, VALUE col1, VALUE col2, VALUE weight)
{
  return COLOR2VALUE(fl_color_average(
    VALUE2COLOR(col1), VALUE2COLOR(col2), NUM2DBL(weight)
  ));
};

#ifdef HAVE_FL_INACTIVE
static VALUE rb_fltk_inactive(VALUE self, VALUE arg1)
{
  return COLOR2VALUE(fl_inactive(VALUE2COLOR(arg1)));
};
#endif

#ifdef HAVE_FL_CONTRAST
static VALUE rb_fltk_contrast(VALUE self, VALUE arg1, VALUE arg2)
{
  return COLOR2VALUE(fl_contrast(VALUE2COLOR(arg1), VALUE2COLOR(arg2)));
};
#endif

extern "C" void
Init_rubyfltk(void)
{
  rb_id_handle = rb_intern("handle");
  rb_id_draw   = rb_intern("draw");
  rb_id_phandle = rb_intern("_handle");
  rb_id_pdraw   = rb_intern("_draw");
  rb_id_icallback = rb_intern("@_callback_");
  rb_id_icallback_data = rb_intern("@_callback_data_");
  rb_id_iitems = rb_intern("@_items_");
  rb_id_ilink = rb_intern("@_link_");
  rb_id_call = rb_intern("call");
  rb_id_clear = rb_intern("clear");
  rb_id_delete = rb_intern("delete");
  rb_id_resize = rb_intern("resize");
  rb_id_flush  = rb_intern("flush");
  rb_id_show   = rb_intern("show");
  rb_id_hide   = rb_intern("hide");
  rb_id_draw_overlay = rb_intern("draw_overlay");
  rb_id_index      = rb_intern("index");
  rb_id_delete_if  = rb_intern("delete_if");

  rb_fltk_thread_schedule_time = 0.0;

  rb_mFLTK = rb_define_module("FLTK");
  rb_define_const(rb_mKernel, "Fltk", rb_mFLTK);

  rb_eFLError = rb_define_class_under(rb_mFLTK, "FLError", rb_eStandardError);

  rb_fltk_objmap = rb_hash_new();
  rb_fltk_fdmap  = rb_hash_new();
  rb_fltk_idle_callbacks    = rb_ary_new();
  rb_fltk_timeout_callbacks = rb_ary_new();
  rb_fltk_check_callbacks   = rb_ary_new(); 
  rb_fltk_handlers  = rb_ary_new();
  rb_define_const(rb_mFLTK, "ObjectMap", rb_fltk_objmap);
  rb_define_const(rb_mFLTK, "FDMap", rb_fltk_fdmap);
  rb_define_const(rb_mFLTK, "IdleCallbacks",    rb_fltk_idle_callbacks);
  rb_define_const(rb_mFLTK, "TimeoutCallbacks", rb_fltk_timeout_callbacks);
  rb_define_const(rb_mFLTK, "CheckCallbacks",   rb_fltk_check_callbacks);
  rb_define_const(rb_mFLTK, "Handlers", rb_fltk_handlers);

  Fl::add_handler(rb_fltk_internal_handler);

  Init_flenum(); /* This should be called before all widgets. */
  Init_flcolor();
  Init_flwidget();
  Init_flbox();
  Init_flbutton();
  Init_flbrowser();
  Init_flchart();
  Init_flclock();
  Init_flgroup();
  Init_flcolorchooser();
  Init_flpack();
  Init_flscroll();
  Init_fltabs();
  Init_fltile();
  Init_flwindow();
  Init_flinput();
  Init_flmenu();
  Init_flpositioner();
  Init_flvaluator();
  Init_flpixmap();
  Init_flbitmap();
  Init_flimage();

#ifdef HAVE_FL_FL_HELP_VIEW_H
  Init_flhelp();
#endif
#ifdef HAVE_FL_FL_FILE_CHOOSER_H
  Init_flfile();
#endif
#ifdef HAVE_FL_FL_TEXT_BUFFER_H
  Init_fltext();
#endif
#ifdef HAVE_FL_FL_WIZARD_H
  Init_flwizard();
#endif
#ifdef HAVE_FL_FL_TOOLTIP_H
  Init_fltooltip();
#endif
#ifdef HAVE_FL_FL_PROGRESS_H
  Init_flprogress();
#endif
#ifdef HAVE_FL_FL_SPINNER_H
  Init_flspinner();
#endif

  DEF_MODULE_FUNC(enable_thread, -1);
  DEF_MODULE_FUNC(disable_thread, 0);
  DEF_MODULE_FUNC_EQ(thread_freq, -1);
  DEF_MODULE_FUNC(add_fd, -1);
  DEF_MODULE_FUNC(remove_fd, 1);
  DEF_MODULE_FUNC(add_idle, -1);
  DEF_MODULE_FUNC(remove_idle, -1);
  DEF_MODULE_FUNC(add_timeout, -1);
  DEF_MODULE_FUNC(repeat_timeout, -1);
  DEF_MODULE_FUNC(remove_timeout, -1);
  DEF_MODULE_FUNC(add_check, -1); 
  DEF_MODULE_FUNC(remove_check, -1); 
  DEF_MODULE_FUNC(add_handler, -1);
  DEF_MODULE_FUNC(run,0);
  DEF_MODULE_FUNC(check,0);
  DEF_MODULE_FUNC(set_font,2);
  DEF_MODULE_FUNC(set_fonts,-1);
  DEF_MODULE_FUNC(event,0);
  DEF_MODULE_FUNC(event_x,0);
  DEF_MODULE_FUNC(event_y,0);
  DEF_MODULE_FUNC(event_x_root,0);
  DEF_MODULE_FUNC(event_y_root,0);
#ifdef HAVE_FL_EVENT_DX
  DEF_MODULE_FUNC(event_dx,0);
  DEF_MODULE_FUNC(event_dy,0);
#endif
  DEF_MODULE_FUNC(event_length,0);
  DEF_MODULE_FUNC(event_text,0);
  DEF_MODULE_FUNC_Q(event_shift,0);
  DEF_MODULE_FUNC_Q(event_ctrl,0);
  DEF_MODULE_FUNC_Q(event_alt,0);
  DEF_MODULE_FUNC(event_button,0);
  DEF_MODULE_FUNC_Q(event_button1,0);
  DEF_MODULE_FUNC_Q(event_button2,0);
  DEF_MODULE_FUNC_Q(event_button3,0);
  DEF_MODULE_FUNC(event_key,-1);
  DEF_MODULE_FUNC(event_clicks,-1);
  DEF_MODULE_FUNC(event_inside,-1);
  DEF_MODULE_FUNC(event_is_click,-1);
  DEF_MODULE_FUNC(event_state,-1);
  DEF_MODULE_FUNC(get_key,1);
  DEF_MODULE_FUNC(w,0);
  DEF_MODULE_FUNC(h,0);

  DEF_MODULE_FUNC(get_color,1);
  DEF_MODULE_FUNC(get_color_rgb,1);
  DEF_MODULE_FUNC(set_color,-1);
  DEF_MODULE_FUNC(foreground,3);
  DEF_MODULE_FUNC(background,3);
  DEF_MODULE_FUNC(background2,3);
  DEF_MODULE_FUNC(color_average,3);
#ifdef HAVE_FL_INACTIVE
  DEF_MODULE_FUNC(inactive,1);
#endif
#ifdef HAVE_FL_CONTRAST
  DEF_MODULE_FUNC(contrast,2);
#endif
  DEF_MODULE_FUNC(lighter,1);
  DEF_MODULE_FUNC(darker,1);

  DEF_UNDERSCORED_MODULE_FUNC(push_clip,4);
  DEF_UNDERSCORED_MODULE_FUNC(push_no_clip,0);
  DEF_MODULE_FUNC(pop_clip,0);
  DEF_MODULE_FUNC_Q(not_clipped,4);
  DEF_MODULE_FUNC(clip_box,4);
  DEF_MODULE_FUNC(paste,1);
  DEF_UNDERSCORED_MODULE_FUNC(line_style,-1);
  DEF_MODULE_FUNC(draw,-1);
  DEF_MODULE_FUNC_EQ(font,-1);
  DEF_MODULE_FUNC_EQ(size,-1);
  DEF_MODULE_FUNC(width,-1);
  DEF_MODULE_FUNC(height,0);
  DEF_MODULE_FUNC(descent,0);
  DEF_UNDERSCORED_MODULE_FUNC(color,-1);
  DEF_MODULE_FUNC(line,-1);
  DEF_MODULE_FUNC(xyline,-1);
  DEF_MODULE_FUNC(yxline,-1);
  DEF_MODULE_FUNC(rect,4);
  DEF_MODULE_FUNC(rectf,-1);
  DEF_MODULE_FUNC(pie,6);
  DEF_MODULE_FUNC(arc,6);
  DEF_MODULE_FUNC(point,2);
  DEF_MODULE_FUNC(circle,3);
  DEF_MODULE_FUNC(begin_loop,0);
  DEF_MODULE_FUNC(end_loop,0);
  DEF_MODULE_FUNC(begin_line,0);
  DEF_MODULE_FUNC(end_line,0);
  DEF_MODULE_FUNC(begin_polygon,0);
  DEF_MODULE_FUNC(end_polygon,0);
  DEF_MODULE_FUNC(begin_complex_polygon,0);
  DEF_MODULE_FUNC(end_complex_polygon,0);
  DEF_MODULE_FUNC(gap,0);
  DEF_MODULE_FUNC(vertex,2);
  DEF_MODULE_FUNC(draw_pixmap,-1);
  DEF_MODULE_FUNC(measure_pixmap,1);
#ifdef HAVE_FL_READ_IMAGE
  DEF_MODULE_FUNC(read_image,-1);
#endif

  DEF_MODULE_FORMAT_FUNC(message);
  DEF_MODULE_FORMAT_FUNC(alert);
  DEF_UNDERSCORED_MODULE_FORMAT_FUNC(ask);
  DEF_MODULE_FORMAT_FUNC_ALT(choice);
  DEF_MODULE_FORMAT_FUNC_ALT(input);
  DEF_MODULE_FORMAT_FUNC_ALT(password);
  DEF_MODULE_FUNC(beep,-1);
  DEF_MODULE_FUNC(rgb_color,-1);

  DEF_MODULE_FUNC(file_chooser,-1);
  DEF_MODULE_FUNC(dir_chooser,-1);
  DEF_MODULE_FUNC(file_chooser_callback,-1);
  DEF_MODULE_FUNC(wait,-1);
  DEF_MODULE_FUNC(flush,0);
  DEF_MODULE_FUNC(focus,-1);
  DEF_MODULE_FUNC(visual,1);
  DEF_MODULE_FUNC(x,0);
  DEF_MODULE_FUNC(y,0);
  DEF_MODULE_FUNC(w,0);
  DEF_MODULE_FUNC(h,0);

#if !defined(WIN32)
  DEF_MODULE_FUNC(set_xid,2);
#endif
  DEF_MODULE_FUNC(xid,1);
  DEF_MODULE_FUNC(find,1);

#ifdef USE_GL
  rb_mFLTKGL = rb_define_module_under(rb_mFLTK, "GL");
  DEF_GL_MODULE_FUNC(color,1);
#endif
#ifdef USE_UTF8
  rb_define_const(rb_mFLTK, "UTF8", Qtrue);
  DEF_MODULE_FUNC(utflen,1);
  DEF_MODULE_FUNC(utf_nb_char,1);
#endif

#if defined(USE_SETLOCALE) || defined(USE_XSETLOCALE)
#if !defined(USE_XLOCALE)
  /* here is a test code */
  if( !setlocale(LC_CTYPE, getenv("LANG")) ){
    if( !setlocale(LC_CTYPE, "") ){
      /* raise error? */
    };
  };
#endif
#endif
};
