/* -*- C++ -*-
 * $Id: fltk.h,v 1.148 2006/12/31 11:30:36 jeremy_henty Exp $
 */

#ifndef RBFLTK_H
#define RBFLTK_H

extern "C" {
#include "ruby.h"
void rb_gc_mark(VALUE);
}

#include <FL/Fl.H>
#include <FL/fl_draw.H>
#include <FL/Enumerations.H>

#ifdef select
# undef select
#endif

#define FLTK_VERSION (FL_MAJOR_VERSION * 10000 + \
                      FL_MINOR_VERSION * 100 + \
                      FL_PATCH_VERSION)
#define FLTK_VERSION_CODE(x,y,z) (x * 10000 + y * 100 + z)

/* Ruby 1.6.x compatibility */
#ifndef HAVE_RB_BLOCK_PROC
# define rb_block_proc rb_f_lambda
#endif

#ifndef LONG2NUM
# define LONG2NUM(x) INT2NUM((int)x)
#endif
#ifndef NUM2LONG
# define NUM2LONG(x) ((long)NUM2INT(x))
#endif
#ifndef UINT2NUM
# define UINT2NUM(x) INT2NUM(x)
#endif
#ifndef NUM2UINT
# define NUM2UINT(x) ((unsigned int)NUM2INT(x))
#endif
#ifndef ULONG2NUM
# define ULONG2NUM(x) UINT2NUM(x)
#endif
#ifndef NUM2ULONG
# define NUM2ULONG(x) ((unsigned long)NUM2UINT(x))
#endif

#define NUM2UCHAR(x) \
((uchar)NUM2UINT(x))

#define UCHAR2NUM(x) \
UINT2NUM((uint)(x))

#ifndef STR2CSTR
static char *rb_fltk_str2cstr(VALUE str)
{
  Check_Type(str, T_STRING);
  return RSTRING(x)->ptr;
}
# define STR2CSTR(x) rb_fltk_str2cstr(x)
#endif

#define STR2CPTR(x) ((x == Qnil) ? NULL : STR2CSTR(x))
#define INT2BOOL(x) (x ? Qtrue : Qfalse)
#define BOOL2INT(x) (x == Qtrue ? 1 : 0)
#define CPTR2STR(x) (x ? rb_tainted_str_new2(x) : Qnil)

#define CPTR2VAL(x) ((((void*)(x))==NULL) ? Qnil : ((((VALUE)(x))==Qnil) ? Qfalse : ((VALUE)(x))))
#define VAL2CPTR(x) (((x) == Qfalse) ? ((void*)Qnil) : (((x) == Qnil) ? NULL : ((void*)(x))))


/* old RUBY_METHOD_FUNC() definition doesn't match to prototypes in those days. */
#ifndef ANYARGS
#undef  RUBY_METHOD_FUNC
#define RUBY_METHOD_FUNC(func) ((VALUE (*)())func)
#endif

/* utility functions */
VALUE rb_ary_insert(VALUE ary, VALUE idx, VALUE data);


#define RBFL_THREAD_SCHEDULE_TIME 0.1

extern VALUE rb_mFLTK;
extern VALUE rb_mFLTKConst;
extern VALUE rb_mFLTKGL;
extern VALUE rb_eFLError;

extern ID rb_id_handle;
extern ID rb_id_phandle;
extern ID rb_id_draw;
extern ID rb_id_pdraw;
extern ID rb_id_call;
extern ID rb_id_icallback;
extern ID rb_id_icallback_data;
extern ID rb_id_iitems;
extern ID rb_id_ilink;
extern ID rb_id_call;
extern ID rb_id_clear;
extern ID rb_id_resize;
extern ID rb_id_flush;
extern ID rb_id_show;
extern ID rb_id_hide;
extern ID rb_id_draw_overlay;

extern "C" {
extern VALUE rb_fltk_objmap_aref(Fl_Widget *);
extern void rb_fltk_objmap_aset(Fl_Widget *, VALUE);
extern void rb_fltk_objmap_delete(Fl_Widget *);
}
extern int *iary2cary(VALUE val, int last = 0);
extern float *fary2cary(VALUE val, float last = 0.0);
extern VALUE cary2iary(int *is, int last = 0);
extern VALUE cary2iary_n(int *is, int len);
extern VALUE cary2fary_n(float *fs, int len);

#define RBFL_CALLER_HANDLE(Fl_Class) \
  virtual int handle(int e){ \
    VALUE ret, rb_object; \
    rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      ret = rb_funcall(rb_object, rb_id_handle, 1, INT2NUM(e)); \
      if( rb_fl_destroyed(rb_object) == Qtrue ) return 1; \
      switch( TYPE(ret) ){ \
      case T_FIXNUM: \
	return FIX2INT(ret); \
      case T_TRUE: \
	return 1; \
      default: \
	return 0; \
      }; \
    } \
    else{ \
      return 0; \
    }; \
  }; \
  virtual int call_parent_handle(int e){ \
    return Fl_Class::handle(e); \
  }; \

#define RBFL_CALLER_DRAW(Fl_Class) \
  virtual void draw(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_draw, 0); \
    }; \
  }; \
  virtual void call_parent_draw(){ \
    Fl_Class::draw(); \
  }; \

#define RBFL_CALLER_RESIZE(Fl_Class) \
  virtual void resize(int x, int y, int w, int h){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_resize, 4, \
            INT2NUM(x), INT2NUM(y), INT2NUM(w), INT2NUM(h)); \
    }; \
  }; \
  virtual void call_parent_resize(int x, int y, int w, int h){ \
    Fl_Class::resize(x,y,w,h); \
  };

#define RBFL_CALLER_FLUSH(Fl_Class) \
  virtual void flush(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_flush, 0); \
    }; \
  }; \
  virtual void call_parent_flush(){ \
    Fl_Class::flush(); \
  };

#define RBFL_CALLER_SHOW(Fl_Class) \
  virtual void show(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_show, 0); \
    }; \
  }; \
  virtual void call_parent_show(){ \
    Fl_Class::show(); \
  };

#define RBFL_CALLER_HIDE(Fl_Class) \
  virtual void hide(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_hide, 0); \
    }; \
  }; \
  virtual void call_parent_hide(){ \
    Fl_Class::hide(); \
  };

#define RBFL_CALLER_DRAW_OVERLAY(Fl_Class) \
  virtual void draw_overlay(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_draw_overlay, 0); \
    }; \
  }; \
  virtual void call_parent_draw_overlay(){ /* empty */ };

#define RBFL_CALLERS(Fl_Class) \
  RBFL_CALLER_HANDLE(Fl_Class) \
  RBFL_CALLER_DRAW(Fl_Class) \
  RBFL_CALLER_RESIZE(Fl_Class)

#define RBFL_HELP_CALLERS(Fl_Class) \
  RBFL_CALLER_RESIZE(Fl_Class) \

#define RBFL_WIDGET_CALLERS(Fl_Class) \
  RBFL_CALLER_HANDLE(Fl_Class) \
  virtual void draw(){ \
    VALUE rb_object = rb_fltk_objmap_aref(this); \
    if( rb_object != Qnil ){ \
      rb_funcall(rb_object, rb_id_draw, 0); \
    }; \
  }; \
  virtual void call_parent_draw(){ /** do nothing **/ }; \
  RBFL_CALLER_RESIZE(Fl_Class) \

#define RBFL_WINDOW_CALLERS(Fl_Class) \
  RBFL_CALLERS(Fl_Class) \
  RBFL_CALLER_FLUSH(Fl_Class) \
  RBFL_CALLER_SHOW(Fl_Class) \
  RBFL_CALLER_HIDE(Fl_Class)

#define RBFL_WIZARD_CALLERS(Fl_Class) \
  RBFL_CALLER_HANDLE(Fl_Class) \
  RBFL_CALLER_RESIZE(Fl_Class)

#define RBFL_OVERLAY_WINDOW_CALLERS(Fl_Class) \
  RBFL_WINDOW_CALLERS(Fl_Class) \
  RBFL_CALLER_DRAW_OVERLAY(Fl_Class)


#define RBFL_CLASS(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_WIDGET(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_WIDGET_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_MENU_(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_WIDGET_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_MENU(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_HELP(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_HELP_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_GROUP(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_WIN(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL##klass(int w, int h, const char *label = 0) \
    : Fl_##klass(w,h,label){}; \
  RBFL_WINDOW_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_OVERLAY_WIN(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL##klass(int w, int h, const char *label = 0) \
    : Fl_##klass(w,h,label){}; \
  RBFL_OVERLAY_WINDOW_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_OVERLAY_GLWIN(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL##klass(int w, int h, const char *label = 0) \
    : Fl_##klass(w,h,label){}; \
  RBFL_OVERLAY_WINDOW_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_WIZARD(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(x,y,w,h,label){}; \
  RBFL_WIZARD_CALLERS(Fl_##klass); \
};

#define RBFL_CLASS_TIMER(klass) \
class RBFL##klass : public Fl_##klass \
{ \
public: \
  RBFL##klass(unsigned char c, int x, int y, int w, int h, const char *label = 0) \
    : Fl_##klass(c,x,y,w,h,label){}; \
  RBFL_CALLERS(Fl_##klass); \
};

/*
 * naming policy:
 * rb_FLWidget_method(...) for class FLWidget and Fl_Widget
 * rb_fltk_method(...)   for module Fltk
 */

#define FN_S_NEW(klass) \
VALUE rb_FL##klass##_s_new(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  widget = new RBFL##klass(0, 0, 0, 0); \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_S_ALLOC(klass) \
VALUE rb_FL##klass##_s_alloc(VALUE klass) \
{ \
  VALUE obj; \
  obj = Data_Wrap_Struct(klass, rb_fl_mark, rb_fl_free, 0); \
  return obj; \
}

#define FN_S_NEW_WIN(klass) \
VALUE rb_FL##klass##_s_new(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  widget = new RBFL##klass(0, 0); \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_S_POPUP_WIN(klass) \
VALUE rb_FL##klass##_s_popup(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  widget = new RBFL##klass(0, 0, 0, 0); \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_S_GET_BOOL_SET_INT(klass,method) \
VALUE rb_FL##klass##_s_##method(VALUE self, VALUE arg1) \
{ \
  return INT2BOOL(RBFL##klass::method(NUM2INT(arg1))); \
};

#define FN_INIT(klass) \
VALUE rb_FL##klass##_initialize(int argc, VALUE argv[], VALUE self) \
{ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  VALUE rx,ry,rw,rh,rlabel; \
  switch( rb_scan_args(argc, argv, "05", &rx, &ry, &rw, &rh, &rlabel) ){ \
  case 0: \
    break; \
  case 1: \
    rlabel = rx; \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 2: \
    rw = rx; rh = ry; \
    rx = INT2NUM(widget->x()); ry = INT2NUM(widget->y()); \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    break; \
  case 3: \
    rlabel = rw; rw = rx; rh = ry; \
    rx = INT2NUM(widget->x()); ry = INT2NUM(widget->y()); \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  case 4: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    break; \
  case 5: \
    rb_funcall(self, rb_id_resize, 4, rx, ry, rw, rh); \
    widget->label(strdup(STR2CSTR(rlabel))); \
    break; \
  }; \
  return Qnil; \
};

#define FN_GROUP_INIT(klass) \
VALUE rb_FL##klass##_initialize(int argc, VALUE argv[], VALUE self) \
{ \
  rb_call_super(argc, argv); \
  if( rb_block_given_p() ){ \
    rb_ensure(RUBY_METHOD_FUNC(rb_yield), self, \
              RUBY_METHOD_FUNC(rb_FLGroup_end), self); \
  }; \
  return Qnil; \
};

#define FN_INIT2(klass, size, extcode) \
VALUE rb_FL##klass##_initialize(int argc, VALUE argv[], VALUE self) \
{ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  rb_call_super(argc, argv); \
  extcode; \
  return Qnil; \
};

#define FN_S_NEW_TIMER(klass) \
VALUE rb_FL##klass##_s_new(int argc, VALUE argv[], VALUE klass) \
{ \
  VALUE obj; \
  RBFL##klass *widget; \
  VALUE rc,rx,ry,rw,rh,rlabel; \
  switch( rb_scan_args(argc, argv, "06", &rc, &rx, &ry, &rw, &rh, &rlabel) ){ \
  case 0: \
    widget = new RBFL##klass(0, 0, 0, 0, 0); \
    break; \
  case 1: \
    widget = new RBFL##klass(rc, 0, 0, 0, 0); \
    break; \
  case 2: \
    rlabel = rx; \
    widget = new RBFL##klass(rc, 0, 0, 0, 0, strdup(STR2CSTR(rlabel))); \
    break; \
  case 3: \
    rw = rx; rh = ry; \
    widget = new RBFL##klass(rc, 0, 0, NUM2INT(rw), NUM2INT(rh)); \
    break; \
  case 4: \
    rlabel = rw; rw = rx; rh = ry; \
    widget = new RBFL##klass(rc, 0, 0, NUM2INT(rw), NUM2INT(rh), \
                             strdup(STR2CSTR(rlabel))); \
    break; \
  case 5: \
    widget = new RBFL##klass(rc, \
                             NUM2INT(rx), NUM2INT(ry), \
			     NUM2INT(rw), NUM2INT(rh)); \
    break; \
  case 6: \
    widget = new RBFL##klass(rc, \
                             NUM2INT(rx), NUM2INT(ry), \
			     NUM2INT(rw), NUM2INT(rh), \
                             strdup(STR2CSTR(rlabel))); \
    break; \
  }; \
  obj = rb_fl_new(klass, widget); \
  rb_obj_call_init(obj, argc, argv); \
  return obj; \
};

#define FN_HANDLE(klass) \
VALUE rb_FL##klass##_##handle(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2BOOL(widget->call_parent_handle(NUM2INT(arg1))); \
};


#define FN_RESIZE(klass) \
VALUE rb_FL##klass##_##resize(VALUE self, VALUE x, VALUE y, VALUE w, VALUE h){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  widget->call_parent_resize(NUM2INT(x), NUM2INT(y), \
                     NUM2INT(w), NUM2INT(h)); \
  return Qnil; \
};

#define FN_FLUSH(klass) FN_CALL_RENAME(klass, flush, call_parent_flush)
#define FN_SHOW(klass)  FN_CALL_RENAME(klass, show, call_parent_show)
#define FN_HIDE(klass)  FN_CALL_RENAME(klass, hide, call_parent_hide)
#define FN_DRAW(klass) FN_CALL_RENAME(klass, draw, call_parent_draw)
#define FN_DRAW_OVERLAY(klass) FN_CALL_RENAME(klass, draw_overlay, call_parent_draw_overlay)

#define WIDGET_FNS(klass) \
  static FN_HANDLE(klass) \
  static FN_DRAW(klass) \
  static FN_RESIZE(klass)

#define HELP_FNS(klass) \
  static FN_RESIZE(klass)

#define WINDOW_FNS(klass) \
  WIDGET_FNS(klass) \
  static FN_FLUSH(klass) \
  static FN_SHOW(klass) \
  static FN_HIDE(klass)

#define OVERLAY_WINDOW_FNS(klass) \
  WINDOW_FNS(klass) \
  static FN_DRAW_OVERLAY(klass)

#define WIZARD_FNS(klass) \
  static FN_HANDLE(klass) \
  static FN_RESIZE(klass)

#define FN_MENU_ADD(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  char* str = (arg1 == Qnil) ? NULL : STR2CSTR(arg1); \
  widget->method(str,0,0); \
  return Qnil; \
};

#define FN_CALL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  widget->method(); \
  return Qnil; \
};

#define FN_CALL_RENAME(klass,rb_method, cpp_method) \
VALUE rb_FL##klass##_##rb_method(VALUE self){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  widget->cpp_method(); \
  return Qnil; \
};

#define FN_GET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return INT2NUM((int)(((RBFL##klass*)widget)->method())); \
};

#define FN_GET_BOOL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  int result = widget->method(); \
  if(result == 0) \
	  return Qfalse; \
  return Qtrue; \
};

#define FN_GET_BOOL_SET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2BOOL(widget->method((int)NUM2INT(arg1))); \
};

#define FN_GET_INT_SET_ENUM(klass,type,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
    return INT2NUM(widget->method((type)NUM2INT(arg1))); \
};

#define FN_GET_INT_SET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2NUM(widget->method((int)NUM2INT(arg1))); \
};

#define FN_GET_INT_SET_INT4(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2, \
                                          VALUE arg3, VALUE arg4){ \
  RBFL##klass *widget = (RBFL##klass *)rb_to_fl(self); \
  return INT2NUM(widget->method((int)NUM2INT(arg1),(int)NUM2INT(arg2), \
                                (int)NUM2INT(arg3),(int)NUM2INT(arg4))); \
};

#define FN_GET_INT_SET_INT_OPTSTR(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1, arg2; \
  switch( rb_scan_args(argc, argv, "11", &arg1, &arg2) ){ \
  case 1: \
    return INT2NUM(widget->method((int)NUM2INT(arg1))); \
  case 2: \
    return INT2NUM(widget->method((int)NUM2INT(arg1),STR2CSTR(arg2))); \
  }; \
};


#define FN_GETSET_ENUM(klass,type,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return INT2NUM(widget->method()); \
  case 1: \
    widget->method((type)NUM2INT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_SET_ENUM(klass,type1,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method((type1)NUM2INT(arg1)); \
  return Qnil; \
};

#define FN_SET_ENUM2(klass,type1,type2,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method((type1)NUM2INT(arg1), \
	         (type2)NUM2INT(arg2)); \
  return Qnil; \
};

#define FN_SET_ENUM4(klass,type1,type2,type3,type4,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method((type1)NUM2INT(arg1), \
	         (type2)NUM2INT(arg2), \
		 (type3)NUM2INT(arg3), \
		 (type4)NUM2INT(arg4)); \
  return Qnil; \
};

#define FN_SET_INT(klass,method)  FN_SET_ENUM(klass,int,method)
#define FN_SET_INT2(klass,method) FN_SET_ENUM2(klass,int,int,method)
#define FN_SET_INT4(klass,method) FN_SET_ENUM4(klass,int,int,int,int,method)

#define FN_SET_ENUM_OPTENUM(klass, ty1, ty2, method) \
  VALUE rb_FL##klass##_##method(VALUE argc, VALUE argv[], VALUE self) { \
    RBFL##klass* widget = (RBFL##klass*)rb_to_fl(self); \
    VALUE arg1, arg2; \
    switch( rb_scan_args(argc, argv, "11", &arg1, &arg2) ){ \
      case 1: \
        widget->method((ty1)NUM2INT(arg1)); \
        break; \
      case 2: \
        widget->method((ty1)NUM2INT(arg1), (ty2)NUM2INT(arg2)); \
	break; \
    } \
    return Qnil; \
  };

#define FN_SET_INT_OPTINT(klass, method) \
        FN_SET_ENUM_OPTENUM(klass,int,int,method)
#define FN_SET_INT_OPTINT2(klass, method) \
        FN_SET_ENUM_OPTENUM(klass,int,int,int,method)

#define FN_SET_INT2_OPTINT3(klass, method) \
  VALUE rb_FL##klass##_##method(VALUE argc, VALUE argv[], VALUE self) { \
    RBFL##klass* widget = (RBFL##klass*)rb_to_fl(self); \
    VALUE arg1, arg2, arg3, arg4, arg5; \
    switch( rb_scan_args(argc, argv, "23", &arg1, &arg2, &arg3, &arg4, &arg5) ){ \
      case 2: \
        widget->method((int)NUM2INT(arg1), (int)NUM2INT(arg2)); \
		break; \
      case 3: \
        widget->method((int)NUM2INT(arg1), (int)NUM2INT(arg2), (int)NUM2INT(arg3)); \
		break; \
      case 4: \
        widget->method((int)NUM2INT(arg1), (int)NUM2INT(arg2), \
                       (int)NUM2INT(arg3), (int)NUM2INT(arg4)); \
		break; \
      case 5: \
        widget->method((int)NUM2INT(arg1), (int)NUM2INT(arg2), (int)NUM2INT(arg3), \
                       (int)NUM2INT(arg4), (int)NUM2INT(arg5)); \
		break; \
    } \
    return Qnil; \
  };

#define FN_SET_INT_STR(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  RBFL##klass* widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method((int)NUM2INT(arg1), STR2CSTR(arg2)); \
  return Qnil; \
};

#define FN_SET_INT2_STR(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3){ \
  RBFL##klass* widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method((int)NUM2INT(arg1), (int)NUM2INT(arg2), STR2CSTR(arg3)); \
  return Qnil; \
};

#define FN_GETSET_BOOL(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return INT2BOOL(((RBFL##klass*)widget)->method()); \
  case 1: \
    ((RBFL##klass*)widget)->method(BOOL2INT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_GETSET_INT(klass,method) FN_GETSET_ENUM(klass,int,method)

#define FN_GETSET_UINT(klass,type,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return UINT2NUM(widget->method()); \
  case 1: \
    widget->method((type)NUM2UINT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_GET_DBL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return rb_float_new(((RBFL##klass*)widget)->method()); \
};

#define FN_GETSET_DBL(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_float_new(((RBFL##klass*)widget)->method()); \
  case 1: \
    arg1 = rb_Float(arg1); \
    ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value); \
  return Qnil; \
  }; \
};

#define FN_SET_DBL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  Fl_Widget *widget = rb_to_fl(self); \
  arg1 = rb_Float(arg1); \
  ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value); \
  return Qnil; \
};

#define FN_SET_DBL2(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  Fl_Widget *widget = rb_to_fl(self); \
  arg1 = rb_Float(arg1); arg2 = rb_Float(arg2); \
  ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value, RFLOAT(arg2)->value); \
  return Qnil; \
};

#define FN_SET_DBL3(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3){ \
  Fl_Widget *widget = rb_to_fl(self); \
  arg1 = rb_Float(arg1); arg2 = rb_Float(arg2); arg3 = rb_Float(arg3);\
  ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value, RFLOAT(arg2)->value, RFLOAT(arg3)->value); \
  return Qnil; \
};

#define FN_GET_DBL_SET_DBL(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  Fl_Widget *widget = rb_to_fl(self); \
  double ret; \
  arg1 = rb_Float(arg1); \
  ret = ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value); \
  return rb_float_new(ret); \
};

#define FN_GET_DBL_SET_DBL_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  Fl_Widget *widget = rb_to_fl(self); \
  double ret; \
  arg1 = rb_Float(arg1); \
  ret = ((RBFL##klass*)widget)->method(RFLOAT(arg1)->value, (int)NUM2INT(arg2)); \
  return rb_float_new(ret); \
};

#define FN_GET_STR(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return rb_tainted_str_new2(((RBFL##klass*)widget)->method()); \
};

#define FN_GET_STR_SET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  return rb_tainted_str_new2(widget->method((int)NUM2INT(arg1))); \
};

#define FN_GET_STR_SET_INT2(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  return rb_tainted_str_new2(widget->method((int)NUM2INT(arg1), (int)NUM2INT(arg2))); \
};

#define FN_GET_STR_SET_OPTINT(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_tainted_str_new2(((RBFL##klass*)widget)->method()); \
  case 1: \
    return rb_tainted_str_new2(((RBFL##klass*)widget)->method((int)NUM2INT(arg1))); \
  }; \
};

#define FN_SET_STR(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE str){ \
  Fl_Widget *widget = rb_to_fl(self); \
  ((RBFL##klass*)widget)->method(STR2CSTR(str)); \
  return Qnil; \
};

#define FN_GET_INT_SET_STR(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE str){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return INT2NUM((int)(((RBFL##klass*)widget)->method(STR2CSTR(str)))); \
};

#define FN_GET_INT_SET_STR_OPTINT(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE str, i; \
  switch( rb_scan_args(argc, argv, "11", &str, &i) ){ \
  case 1: \
    return INT2NUM((int)(widget->method(STR2CSTR(str)))); \
  case 2: \
    return INT2NUM((int)(widget->method(STR2CSTR(str),NUM2INT(i)))); \
  } \
};

#define FN_GET_BOOL_SET_STR_OPTBOOL(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE str, i; \
  switch( rb_scan_args(argc, argv, "11", &str, &i) ){ \
  case 1: \
    return INT2BOOL((int)(widget->method(STR2CSTR(str)))); \
  case 2: \
    return INT2BOOL((int)(widget->method(STR2CSTR(str),BOOL2INT(i)))); \
  } \
};

#define FN_GETSET_STR(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass*widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  char *str; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    str = (char*)(widget->method()); \
    return str ? rb_tainted_str_new2(str) : Qnil; \
  case 1: \
    str = (char*)STR2CPTR(arg1); \
    widget->method(str); \
    return Qnil; \
  }; \
};

#define FN_GETSET_STR_COPY(klass,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass*widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  char *str; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    str = (char*)(widget->method()); \
    return str ? rb_tainted_str_new2(str) : Qnil; \
  case 1: \
    str = (char*)(widget->method()); \
    if( str ) free(str); \
    str = (arg1 == Qnil) ? NULL : strdup((char*)STR2CSTR(arg1)); \
    widget->method(str); \
    return Qnil; \
  }; \
};

#define FN_SET_STR_COPY(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass*widget = (RBFL##klass*)rb_to_fl(self); \
  char* str = (arg1 == Qnil) ? NULL : STR2CSTR(arg1); \
  widget->method(str); \
  return Qnil; \
};

#define FN_GET_VALUE(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  return (VALUE)(widget->method()); \
};

#define FN_SET_VALUE(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method((void*)arg1); \
  return Qnil; \
};

#define FN_SET_VALUE2(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method((void*)arg1, (void*)arg2); \
  return Qnil; \
};

#define FN_GET_VALUE_SET_INT(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE arg1){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  return (VALUE)(widget->method((int)NUM2INT(arg1))); \
};

#define FN_GETSET_VALUE(klass,type,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  VALUE obj; \
  switch( rb_scan_args(argc, argv, "01", &obj) ){ \
  case 0: \
    return (VALUE)(((RBFL##klass*)widget)->method()); \
  case 1: \
    ((RBFL##klass*)widget)->method((type)obj); \
    return Qnil; \
  }; \
};

#define FN_GET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self){ \
  Fl_Widget *widget = rb_to_fl(self); \
  return rb_fltk_objmap_aref((Fl_Widget*)((RBFL##klass*)widget)->method()); \
};

#define FN_S_GET_WIDGET(klass,method) \
VALUE rb_FL##klass##_s_##method(VALUE self){ \
  return rb_fltk_objmap_aref((Fl_Widget*)(RBFL##klass::method())); \
};

#define FN_GETSET_OBJECT(klass,type,method) \
VALUE rb_FL##klass##_##method(int argc, VALUE argv[], VALUE self){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_fltk_objmap_aref((Fl_Widget*)(widget->method())); \
  case 1: \
    widget->method((type*)(rb_to_fl(arg1))); \
    return Qnil; \
  }; \
};

#define FN_GETSET_WIDGET(klass,method) FN_GETSET_OBJECT(klass,Fl_Widget,method)

#define FN_GET_BOOL_SET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  if( widget->method(rb_to_fl(w)) ){ \
    return Qtrue; \
  } \
  else{ \
    return Qfalse; \
  }; \
};

#define FN_GET_INT_SET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  return INT2NUM(widget->method(rb_to_fl(w))); \
};

#define FN_SET_WIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method(rb_to_fl(w)); \
  return Qnil; \
};

#define FN_SET_REFWIDGET(klass,method) \
VALUE rb_FL##klass##_##method(VALUE self, VALUE w){ \
  RBFL##klass *widget = (RBFL##klass*)rb_to_fl(self); \
  widget->method(*(rb_to_fl(w))); \
  return Qnil; \
};



#define FN_MODULE_FL_CALL(method) \
VALUE rb_fltk_##method(VALUE self){ \
  Fl::method(); \
  return Qnil; \
};

#define FN_MODULE_FL_GET_INT(method) \
VALUE rb_fltk_##method(VALUE self){ \
  return INT2NUM(Fl::method()); \
};

#define FN_MODULE_FL_GET_UINT(method) \
VALUE rb_fltk_##method(VALUE self){ \
  return UINT2NUM(Fl::method()); \
};

#define FN_MODULE_FL_GET_BOOL(method) \
VALUE rb_fltk_##method(VALUE self){ \
  return INT2BOOL(Fl::method()); \
};

#define FN_MODULE_FL_GET_UINT_SET_ENUM(etype,method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1){ \
  return UINT2NUM(Fl::method((etype)NUM2INT(arg1)); \
};

#define FN_MODULE_FL_GET_INT_SET_INT(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1){ \
  return INT2NUM(Fl::method((int)NUM2INT(arg1))); \
};

#define FN_MODULE_FL_GET_BOOL_SET_INT(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1){ \
  return INT2BOOL(Fl::method((int)NUM2INT(arg1))); \
};

#define FN_MODULE_FL_GETSET_ENUM(etype,method) \
VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self){ \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return INT2NUM((int)(Fl::method())); \
  case 1: \
    Fl::method((etype)NUM2INT(arg1)); \
    return Qnil; \
  }; \
};

#define FN_MODULE_FL_GETSET_INT(method) FN_MODULE_FL_GETSET_ENUM(int,method)

#define FN_MODULE_FL_GET_STR(method,method2) \
VALUE rb_fltk_##method(VALUE self){ \
  return rb_tainted_str_new(Fl::method(),Fl::method2()); \
};

#define FN_MODULE_FL_GET_STR2(method) \
VALUE rb_fltk_##method(VALUE self){ \
  return rb_tainted_str_new2(Fl::method()); \
};

#define FN_MODULE_FL_SET_STR(method) \
VALUE rb_fltk_##method(VALUE mod, VALUE arg1){ \
  Fl::method(STR2CSTR(arg1)); \
  return Qnil;\
};

#define FN_MODULE_FL_SET_UCHAR3(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3) { \
  Fl::method(NUM2UCHAR(arg1), \
             NUM2UCHAR(arg2), \
             NUM2UCHAR(arg3)); \
  return Qnil; \
};

#define FN_MODULE_FL_GETSET_STR(method) \
VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self){ \
  VALUE arg1; \
  char *str; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_tainted_str_new2(Fl::method()); \
  case 1: \
    str = Fl::method(); \
    if( str ) free(str); \
    Fl::method(strdup(STR2CSTR(arg1))); \
    return Qnil; \
  }; \
};

#define FN_MODULE_FL_GETSET_OBJECT(type,method) \
VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE mod){ \
  VALUE arg1; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    return rb_fltk_objmap_aref((Fl_Widget*)(Fl::method())); \
  case 1: \
    Fl::method((type*)(rb_to_fl(arg1))); \
    return Qnil; \
  }; \
};

#define FN_MODULE_FL_GETSET_WIDGET(method) \
        FN_MODULE_FL_GETSET_OBJECT(Fl_Widget,method)

#define FN_GLOBAL_CALL(method) \
VALUE rb_fltk_##method(VALUE self){ \
  fl_##method(); \
  return Qnil; \
};

#define FN_GL_GLOBAL_CALL(method) \
VALUE rb_fltk_##gl_##method(VALUE self){ \
  gl_##method(); \
  return Qnil; \
};

#define FN_GLOBAL_GET_INT(method) \
VALUE rb_fltk_##method(VALUE self) { \
  return INT2NUM(fl_##method()); \
};

#define FN_GLOBAL_GET_BOOL_SET_INT4(method) \
VALUE rb_fltk_##method(VALUE mod, VALUE x, VALUE y, VALUE w, VALUE h){ \
  return fl_##method(NUM2INT(x),NUM2INT(y),NUM2INT(w),NUM2INT(h)) ? Qtrue : Qfalse; \
};

#define FN_GLOBAL_GETSET_INT(method) \
  VALUE rb_fltk_##method(VALUE argc, VALUE argv[], VALUE self) { \
    VALUE arg1; \
    switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
      case 0: \
        return INT2NUM(fl_##method()); \
      case 1: \
        fl_##method(NUM2INT(arg1)); \
		break; \
    } \
    return Qnil; \
  };

#define FN_GLOBAL_SET_ENUM(method,type) \
  VALUE rb_fltk_##method(VALUE self, VALUE x) { \
    fl_##method((type)NUM2INT(x)); \
    return Qnil; \
  };

#define FN_GL_GLOBAL_SET_ENUM(method,type) \
  VALUE rb_fltk_##gl_##method(VALUE self, VALUE x) { \
    gl_##method((type)NUM2INT(x)); \
    return Qnil; \
  };

#define FN_GLOBAL_SET_INT(method) FN_GLOBALE_SET_ENUM(method,int)
#define FN_GL_GLOBAL_SET_INT(method) FN_GL_GLOBAL_SET_ENUM(method,int)

#define FN_GLOBAL_SET_INT2(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE x, VALUE y) { \
    fl_##method(NUM2INT(x), NUM2INT(y)); \
    return Qnil; \
  };

#define FN_GL_GLOBAL_SET_INT2(method) \
  VALUE rb_fltk_##gl_##method(VALUE self, VALUE x, VALUE y) { \
    gl_##method(NUM2INT(x), NUM2INT(y)); \
    return Qnil; \
  };

#define FN_GLOBAL_SET_INT4_OPTINT2(method) \
  VALUE rb_fltk_##method(VALUE argc, VALUE argv[], VALUE self) { \
    VALUE arg1, arg2, arg3, arg4, arg5, arg6; \
    switch( rb_scan_args(argc, argv, "42", &arg1, &arg2, &arg3, &arg4, &arg5, &arg6) ){ \
      case 4: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4)); \
		break; \
      case 6: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4), \
		           NUM2INT(arg5), NUM2INT(arg6)); \
		break; \
      default: \
	 rb_raise(rb_eArgError, "too few/many arguments"); \
         break; \
    } \
    return Qnil; \
  };

#define FN_GLOBAL_SET_INT4(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE x, VALUE y, VALUE w, VALUE h) { \
    fl_##method(NUM2INT(x), NUM2INT(y), NUM2INT(w), NUM2INT(h)); \
    return Qnil; \
  };

#define FN_GLOBAL_SET_INT4_OPTCHR3(method) \
  VALUE rb_fltk_##method(VALUE argc, VALUE argv[], VALUE self) { \
    VALUE arg1, arg2, arg3, arg4, arg5, arg6, arg7; \
    switch( rb_scan_args(argc, argv, "42", &arg1, &arg2, &arg3, &arg4, \
	             &arg5, &arg6, arg7) ){ \
      case 4: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4)); \
		break; \
      case 7: \
        fl_##method(NUM2INT(arg1), NUM2INT(arg2), NUM2INT(arg3), NUM2INT(arg4), \
		           NUM2CHR(arg5), NUM2CHR(arg6), NUM2CHR(arg7)); \
		break; \
      default: \
	    /* TODO: should raise an error here */ \
	    break; \
    } \
    return Qnil; \
  };

#define FN_GLOBAL_SET_INT4_DBL2(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE x, VALUE y, VALUE w, VALUE h, VALUE a1, VALUE a2) { \
    fl_##method(NUM2INT(x), NUM2INT(y), NUM2INT(w), NUM2INT(h), \
		           NUM2DBL(a1), NUM2DBL(a2)); \
    return Qnil; \
  };

#define FN_GLOBAL_SET_DBL2(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1, VALUE arg2){ \
  fl_##method(NUM2DBL(arg1), NUM2DBL(arg2)); \
  return Qnil; \
}

#define FN_GLOBAL_SET_DBL3(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3){ \
  fl_##method(NUM2DBL(arg1), NUM2DBL(arg2), NUM2DBL(arg3)); \
  return Qnil; \
}

#define FN_GLOBAL_SET_DBL4(method) \
VALUE rb_fltk_##method(VALUE self, VALUE arg1, VALUE arg2, VALUE arg3, VALUE arg4){ \
  fl_##method(NUM2DBL(arg1), NUM2DBL(arg2), NUM2DBL(arg3), NUM2DBL(arg4)); \
  return Qnil; \
}

#define FN_GLOBAL_SET_STR(method) \
  VALUE rb_fltk_##method(VALUE self, VALUE arg1) { \
    char* str = (arg1 == Qnil) ? NULL : STR2CSTR(arg1); \
    fl_##method("%s", str); \
    return Qnil; \
  };

#define FN_GLOBAL_GETSET_STR(method) \
VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE mod){ \
  VALUE arg1; \
  char *str; \
  switch( rb_scan_args(argc, argv, "01", &arg1) ){ \
  case 0: \
    str = (char*)method(); \
    return str ? rb_tainted_str_new2(str) : Qnil; \
  case 1: \
    str = (char*)STR2CPTR(arg1); \
    method(str); \
    return Qnil; \
  }; \
};

#define FN_GLOBAL_SET_FORMAT(method) \
  VALUE rb_fltk_##method##0(VALUE self, VALUE arg) { \
    fl_##method("%s", STR2CSTR(arg)); \
    return Qnil; \
  }; \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE val; \
    char *str; \
    switch( argc ){ \
    case 0: \
      rb_raise(rb_eArgError, \
	       "wrong number of arguments (%d for %d)", \
	       argc, 1); \
      break; \
    default: \
      val = rb_f_sprintf(argc, argv); \
      str = (val == Qnil) ? NULL : STR2CSTR(val); \
      break; \
    }; \
    fl_##method("%s", str); \
    return Qnil; \
  };

#define FN_GLOBAL_GET_BOOL_SET_FORMAT(method) \
  VALUE rb_fltk_##method##0(VALUE self, VALUE arg) { \
    if( fl_##method("%s", STR2CSTR(arg)) ){ \
      return Qtrue; \
    } \
    else{ \
      return Qfalse; \
    }; \
  }; \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE val; \
    char *str; \
    switch( argc ){ \
    case 0: \
      rb_raise(rb_eArgError, \
	       "wrong number of arguments (%d for %d)", \
	       argc, 1); \
      break; \
    default: \
      val = rb_f_sprintf(argc, argv); \
      str = (val == Qnil) ? NULL : STR2CSTR(val); \
      break; \
    }; \
    if( fl_##method("%s", str) ){ \
      return Qtrue; \
    } \
    else{ \
      return Qfalse; \
    }; \
  };

#define FN_GLOBAL_GET_STR_SET_FORMAT_STR(method) \
  VALUE rb_fltk_##method##0(int argc, VALUE argv[], VALUE self) { \
    VALUE arg1, arg2; \
    const char *result; \
    switch( rb_scan_args(argc, argv, "11", &arg1, &arg2) ){ \
    case 1: \
      result = fl_##method("%s", NULL, STR2CSTR(arg1)); \
      break; \
    case 2: \
      result = fl_##method("%s", STR2CPTR(arg2), STR2CSTR(arg1)); \
      break; \
    }; \
    if( result ){ \
      return rb_tainted_str_new2(result); \
    } \
    else{ \
      return Qnil; \
    }; \
  }; \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE msg, *sprintf_args; \
    const char *result; \
    switch( argc ){ \
    case 0: \
      rb_raise(rb_eArgError, \
	       "wrong number of arguments (%d for %d)", \
	       argc, 1); \
    case 1: \
      msg = rb_f_sprintf(1,argv); \
      result = fl_##method("%s", NULL, STR2CSTR(msg)); \
      break; \
    case 2: \
      msg = rb_f_sprintf(1,argv); \
      result = fl_##method("%s", STR2CPTR(argv[1]), STR2CSTR(msg)); \
      break; \
    default: \
      sprintf_args = ALLOCA_N(VALUE,argc-1); \
      sprintf_args[0] = argv[0]; \
      memcpy(sprintf_args+1,argv+2,(argc-2)*sizeof(VALUE)); \
      msg = rb_f_sprintf(argc-1,sprintf_args); \
      result = fl_##method("%s", STR2CPTR(argv[1]), STR2CSTR(msg)); \
      break; \
    }; \
    if( result ){ \
      return rb_tainted_str_new2(result); \
    } \
    else{ \
      return Qnil; \
    }; \
  };

#define FN_GLOBAL_GET_INT_SET_FORMAT_STR3(method) \
  VALUE rb_fltk_##method##0(int argc, VALUE argv[], VALUE self) { \
    VALUE msg  = Qnil; \
    VALUE str1 = Qnil, str2 = Qnil, str3 = Qnil; \
    int ret; \
    switch( rb_scan_args(argc, argv, "13", &msg, &str1, &str2, &str3) ){ \
    case 1: \
      ret = fl_##method("%s", \
			NULL,NULL,NULL,STR2CSTR(msg)); \
      break; \
    case 2: \
      ret = fl_##method("%s", \
			STR2CPTR(str1), \
			NULL,NULL,STR2CSTR(msg)); \
      break; \
    case 3: \
      ret = fl_##method("%s", \
			STR2CPTR(str1), \
			STR2CPTR(str2), \
			NULL,STR2CSTR(msg)); \
      break; \
    case 4: \
      ret = fl_##method("%s", \
			STR2CPTR(str1), \
			STR2CPTR(str2), \
			STR2CPTR(str3), \
			STR2CPTR(msg)); \
      break; \
    }; \
    return INT2NUM(ret); \
  }; \
  VALUE rb_fltk_##method(int argc, VALUE argv[], VALUE self) { \
    VALUE msg, *sprintf_args; \
    int ret; \
    switch( argc ){ \
    case 0: \
      rb_raise(rb_eArgError, \
	       "wrong number of arguments (%d for %d)", \
	       argc, 1); \
    case 1: \
      msg = rb_f_sprintf(1,argv); \
      ret = fl_##method("%s", \
			NULL,NULL,NULL,STR2CSTR(msg)); \
      break; \
    case 2: \
      msg = rb_f_sprintf(1,argv); \
      ret = fl_##method("%s", \
			STR2CPTR(argv[1]), \
			NULL,NULL,STR2CSTR(msg)); \
      break; \
    case 3: \
      msg = rb_f_sprintf(1,argv); \
      ret = fl_##method("%s", \
			STR2CPTR(argv[1]), \
			STR2CPTR(argv[2]), \
			NULL,STR2CSTR(msg)); \
      break; \
    case 4: \
      msg = rb_f_sprintf(1,argv); \
      ret = fl_##method("%s", \
			STR2CPTR(argv[1]), \
			STR2CPTR(argv[2]), \
			STR2CPTR(argv[3]), \
			STR2CPTR(msg)); \
      break; \
    default: \
      sprintf_args = ALLOCA_N(VALUE,argc-3); \
      sprintf_args[0] = argv[0]; \
      memcpy(sprintf_args+1,argv+4,(argc-4)*sizeof(VALUE)); \
      msg = rb_f_sprintf(argc-3,sprintf_args); \
      ret = fl_##method("%s", \
			STR2CPTR(argv[1]), \
			STR2CPTR(argv[2]), \
			STR2CPTR(argv[3]), \
			STR2CPTR(msg)); \
      break; \
    }; \
    return INT2NUM(ret); \
  };

/* wrappers for rb_define_{method,singleton_method,module_function} */
#define RB_DEF_METHOD(klass,method,func,argtype) \
        rb_define_method(klass,method,RUBY_METHOD_FUNC(func), argtype)
#define RB_DEF_S_METHOD(klass,method,func,argtype) \
        rb_define_singleton_method(klass,method,RUBY_METHOD_FUNC(func), argtype)
#define RB_DEF_MODULE_FUNC(module,method,func,argtype) \
        rb_define_module_function(module,method,RUBY_METHOD_FUNC(func), argtype)

#define DEF_S_METHOD(klass,method,argtype) \
        RB_DEF_S_METHOD(rb_cFL##klass, #method, rb_FL##klass##_s_##method, argtype)
#define DEF_METHOD(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype)
#define DEF_METHOD_Q(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #method"?", rb_FL##klass##_##method, argtype)
#define DEF_METHOD_EQ(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #method"=", rb_FL##klass##_##method, argtype)
#define DEF_METHOD_QEQ(klass,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #method, rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #method"=", rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #method"?", rb_FL##klass##_##method, argtype)
#define DEF_EQUALITY_METHODS(klass) \
        DEF_METHOD(klass,hash,0); \
        RB_DEF_METHOD(rb_cFL##klass,"eql?",rb_FL##klass##_##eql,1); \
        RB_DEF_METHOD(rb_cFL##klass,"==",rb_FL##klass##_##equal,1);

#define DEF_S_NEW(klass) \
        DEF_S_METHOD(klass, new, -1)

#define DEF_S_ALLOC(klass) \
        DEF_S_METHOD(klass, allocate, 0)

#define DEF_INIT_METHOD(klass) \
        DEF_METHOD(klass,initialize,-1); rb_enable_super(rb_cFL##klass,"initialize")

#define DEF_PREFIXED_METHOD(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype)
#define DEF_PREFIXED_METHOD_Q(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"?", \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"?", \
		      rb_FL##klass##_##method, argtype)
#define DEF_PREFIXED_METHOD_EQ(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"=", \
		      rb_FL##klass##_##method, argtype); \
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"=", \
		      rb_FL##klass##_##method, argtype)
#define DEF_PREFIXED_METHOD_QEQ(klass,prefix,method,argtype) \
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method, \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method, \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"=", \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"=", \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix "_" #method"?", \
		      rb_FL##klass##_##method, argtype);\
        RB_DEF_METHOD(rb_cFL##klass, #prefix #method"?", \
		      rb_FL##klass##_##method, argtype)

#define DEF_MODULE_FUNC(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype)
#define DEF_MODULE_FUNC_Q(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"?", rb_fltk_##method, argtype)
#define DEF_MODULE_FUNC_EQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"=", rb_fltk_##method, argtype)
#define DEF_MODULE_FUNC_QEQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"?", rb_fltk_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method"=", rb_fltk_##method, argtype)

#define DEF_MODULE_FORMAT_FUNC(method) \
	RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, -1); \
	RB_DEF_MODULE_FUNC(rb_mFLTK, #method "0", rb_fltk_##method##0, 1)

#define DEF_MODULE_FORMAT_FUNC_ALT(method) \
	RB_DEF_MODULE_FUNC(rb_mFLTK, #method, rb_fltk_##method, -1); \
	RB_DEF_MODULE_FUNC(rb_mFLTK, #method "0", rb_fltk_##method##0, -1)

#define DEF_UNDERSCORED_MODULE_FUNC(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method "_", rb_fltk_##method, argtype)

#define DEF_UNDERSCORED_MODULE_FORMAT_FUNC(method) \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method "_", rb_fltk_##method, -1); \
        RB_DEF_MODULE_FUNC(rb_mFLTK, #method "0_", rb_fltk_##method##0, 1)

#define DEF_GL_MODULE_FUNC(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype)
#define DEF_GL_MODULE_FUNC_Q(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"?", rb_fltk_##gl_##method, argtype)
#define DEF_GL_MODULE_FUNC_EQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"=", rb_fltk_##gl_##method, argtype)
#define DEF_GL_MODULE_FUNC_QEQ(method,argtype) \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method, rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"?", rb_fltk_##gl_##method, argtype); \
        RB_DEF_MODULE_FUNC(rb_mFLTKGL, #method"=", rb_fltk_##gl_##method, argtype)

#define DEF_WIDGET_METHODS(klass) \
  DEF_METHOD(klass, handle, 1); \
  DEF_METHOD(klass, draw, 0); \
  DEF_METHOD(klass, resize, 4);

#define DEF_HELP_METHODS(klass) \
  DEF_METHOD(klass, resize, 4);

#define DEF_WINDOW_METHODS(klass) \
  DEF_WIDGET_METHODS(klass); \
  DEF_METHOD(klass, flush, 0); \
  DEF_METHOD(klass, show, 0); \
  DEF_METHOD(klass, hide, 0);

#define DEF_OVERLAY_WINDOW_METHODS(klass) \
  DEF_WINDOW_METHODS(klass); \
  DEF_METHOD(klass, draw_overlay, 0); \

#define DEF_WIZARD_METHODS(klass) \
  DEF_METHOD(klass, handle, 1); \
  DEF_METHOD(klass, resize, 4);

#endif /* RBFLTK_H */
