/* -*- C++ -*-
 * $Id: flwidget.cc,v 1.86 2006/12/31 01:13:58 jeremy_henty Exp $
 */

#include "fltk.h"
#include "flwidget.h"
#include "flcolor.h"
#include "flgroup.h"
#include "flmenu.h"
#include "fltext.h"

VALUE rb_cFLWidget;

class RBFLMarkStack {
public:

  VALUE klass;
  RBFLMarkFunc mark;
  RBFLMarkStack *next;

  RBFLMarkStack(VALUE klass, RBFLMarkFunc mark, RBFLMarkStack *next){
    this->klass = klass;
    this->mark = mark;
    this->next = next;
  };
};

static RBFLMarkStack *rb_fltk_mark_stack;

void rb_fltk_add_mark(VALUE klass, RBFLMarkFunc mark)
{
  rb_fltk_mark_stack = new RBFLMarkStack(klass, mark, rb_fltk_mark_stack);
}

void rb_fltk_append_mark(VALUE klass, RBFLMarkFunc mark)
{
  RBFLMarkStack *p;
  if( rb_fltk_mark_stack ){
    for( p = rb_fltk_mark_stack; p; p = p->next ){
      if( ! p->next ){
	p->next = new RBFLMarkStack(klass, mark, NULL);
	break;
      };
    }
  }
  else{
    rb_fltk_add_mark(klass, mark);
  }
}

static void rb_fl_free(Fl_Widget *widget)
{
  /* I assume that 'delete' is done by Widget#destroy method. */
}


static void rb_fl_mark(Fl_Widget *widget)
{
  VALUE obj;
  RBFLMarkStack *p;

  if( !widget ){
    return;
  };

  obj = fl_to_rb(widget);

  for( p=rb_fltk_mark_stack; p; p=p->next ){
    if( rb_obj_is_kind_of(obj,p->klass) ){
      (p->mark)(obj);
    }
  }
}

VALUE rb_fl_new(VALUE klass, Fl_Widget *widget)
{
  VALUE obj;

  obj = Data_Wrap_Struct(klass, rb_fl_mark, rb_fl_free, widget);
  rb_fltk_objmap_aset(widget, obj);
  rb_ivar_set(obj, rb_id_icallback, Qnil);

  return obj;
}

static void rb_fl_destroy_internal(VALUE obj)
{
  Fl_Widget *widget;

  if( obj == Qnil ) return ;

  widget = rb_to_fl(obj);
  rb_fltk_objmap_delete(widget);
  if( rb_obj_is_kind_of(obj, rb_cFLGroup) ){
    Fl_Widget *child;
    int n = ((RBFLGroup*)widget)->children();
    for( int i = 0; i < n; i++ ){
      child = ((RBFLGroup*)widget)->child(i);
      rb_fl_destroy_internal(fl_to_rb(child));
    };
  };
}

VALUE rb_fl_destroy(VALUE self)
{
  Fl_Widget *widget;

  widget = rb_to_fl(self);
  rb_fl_destroy_internal(self);
  if( widget )
    delete widget;

  DATA_PTR(self) = 0;

  return Qnil;
}

VALUE rb_fl_destroyed(VALUE self)
{
  Fl_Widget *widget;

  widget = (Fl_Widget*)DATA_PTR(self);

  if( !widget ) return Qtrue;
  if( fl_to_rb(widget) == Qnil ){
    return Qtrue;
  }
  else{
    return Qfalse;
  };
}

VALUE rb_fl_valid(VALUE self)
{
  Fl_Widget *widget;

  rb_warn("use 'destroyed?' instead of 'valid?'.");
  return (rb_fl_destroyed(self) == Qtrue) ? Qfalse : Qtrue;
}

VALUE rb_fl_check(VALUE self)
{
  if( rb_fl_destroyed(self) == Qtrue ){
    rb_raise(rb_eFLError, "the widget has been destroyed.");
  };
  return Qnil;
}

VALUE rb_fl_check_widget(VALUE self)
{
  if( !rb_obj_is_kind_of(self, rb_cFLWidget) ){
    rb_raise(rb_eTypeError, "%s is not a FLWidget.",
	     STR2CSTR(rb_funcall(self, rb_intern("inspect"), 0)));
  };
  return Qnil;
}

Fl_Widget *rb_to_fl(VALUE self)
{
  if( self == Qnil ){
    return NULL;
  };
  rb_fl_check(self);
  return (Fl_Widget*)DATA_PTR(self);
}

VALUE fl_to_rb(Fl_Widget *widget)
{
  return rb_fltk_objmap_aref(widget);
}


static void rb_flwidget_callback_func(Fl_Widget *widget, void *userdata)
{
  VALUE obj;
  VALUE callback;

  obj = fl_to_rb(widget);
  if( obj == Qnil ) return ;

  callback = rb_ivar_get(obj, rb_id_icallback);

  if( callback != Qnil ){
    rb_funcall(callback, rb_id_call, 2, obj, CPTR2VAL(userdata));
  };
}

static void rb_flwidget_mark(VALUE obj)
{
  RBFLWidget *w = (RBFLWidget*)rb_to_fl(obj);
  void *userdata = w->user_data();
  if( userdata ){
    rb_gc_mark(CPTR2VAL(userdata));
  }
}

static VALUE rb_FLWidget_callback(int argc, VALUE argv[], VALUE self)
{
  VALUE proc, data;
  Fl_Widget *widget;
  int use_data = 0;

  if( rb_block_given_p() ){
    proc = rb_block_proc();
    switch( rb_scan_args(argc, argv, "01", &data) ){
    case 0:
      use_data = 0;
      break;
    case 1:
      use_data = 1;
      break;
    };
  }
  else{
    switch( rb_scan_args(argc, argv, "11", &proc, &data) ){
    case 1:
      use_data = 0;
      break;
    case 2:
      use_data = 1;
      break;
    };
  };

  rb_ivar_set(self, rb_id_icallback, proc);
  widget = rb_to_fl(self);
  if( use_data ){
    ((RBFLWidget*)widget)->callback((Fl_Callback*)rb_flwidget_callback_func,
				    VAL2CPTR(data));
  }
  else{
    ((RBFLWidget*)widget)->callback((Fl_Callback*)rb_flwidget_callback_func);
  };

  return Qnil;
}

VALUE rb_fl_callback(VALUE self)
{
  VALUE argv[1] = {Qnil};

  return rb_FLWidget_callback(0, argv, self);
}

static VALUE rb_FLWidget_user_data(int argc, VALUE argv[], VALUE self)
{
  Fl_Widget *widget = rb_to_fl(self);
  VALUE data;

  switch( rb_scan_args(argc, argv, "01", &data) ){
  case 0:
    return CPTR2VAL(widget->user_data());
  case 1:
    widget->user_data(VAL2CPTR(data));
    return Qnil;
  };
}

static VALUE rb_FLWidget_do_callback(int argc, VALUE argv[], VALUE self)
{
  Fl_Widget *widget = rb_to_fl(self);
  VALUE w, data;

  switch( rb_scan_args(argc, argv, "02", &w, &data) ){
  case 2:
    widget->do_callback(rb_to_fl(w), VAL2CPTR(data));
    break;
  case 1:
    widget->do_callback(rb_to_fl(w));
    break;
  case 0:
    widget->do_callback();
    break;
  };
  return Qnil;
}

static VALUE rb_FLWidget_damage(int argc, VALUE argv[], VALUE self)
{
  Fl_Widget *widget = rb_to_fl(self);
  VALUE vmask, vx, vy, vw, vh;
  uchar mask = 0;
  int x = 0, y = 0, w = 0, h = 0;

  switch( rb_scan_args(argc, argv, "05", &vmask, &vx, &vy, &vw, &vh) ){
  case 5:
    h = NUM2INT(vh);
  case 4:
    w = NUM2INT(vw);
  case 3:
    y = NUM2INT(vy);
  case 2:
    x = NUM2INT(vx);
    mask = NUM2INT(vmask);
    widget->damage(mask,x,y,w,h);
    return Qnil;
  case 1:
    mask = NUM2INT(vmask);
    widget->damage(mask);
    return Qnil;
  case 0:
    return INT2NUM(widget->damage());
  }
}

static FN_S_NEW(Widget);
static FN_INIT(Widget);
WIDGET_FNS(Widget);
static FN_GETSET_INT(Widget,type);
static FN_SET_INT2(Widget,position);
static FN_SET_INT2(Widget,size);
static FN_GET_WIDGET(Widget,window);
static FN_GET_WIDGET(Widget,parent);
static FN_GETSET_ENUM(Widget,Fl_Boxtype,box);
static FN_GETSET_COLOR2(Widget,color);
static FN_GETSET_COLOR(Widget,selection_color);
static FN_GETSET_STR_COPY(Widget,label);
static FN_GETSET_ENUM(Widget,Fl_Labeltype,labeltype);
static FN_GETSET_ENUM(Widget,Fl_Font,labelfont);
static FN_GETSET_COLOR(Widget,labelcolor);
static FN_GETSET_INT(Widget,labelsize);
static FN_GETSET_ENUM(Widget,Fl_Align,align);
static FN_CALL(Widget,show);
static FN_CALL(Widget,hide);
static FN_CALL(Widget,redraw);
static FN_GETSET_ENUM(Widget,Fl_When,when);
static FN_GET_BOOL(Widget,visible);
static FN_GET_BOOL(Widget,visible_r);
static FN_GET_BOOL(Widget,active);
static FN_GET_BOOL(Widget,active_r);
static FN_CALL(Widget,activate);
static FN_CALL(Widget,deactivate);
static FN_GET_BOOL(Widget,output);
static FN_CALL(Widget,set_output);
static FN_CALL(Widget,clear_output);
static FN_GET_BOOL(Widget,takesevents);
static FN_GET_BOOL_SET_WIDGET(Widget,contains);
static FN_GET_BOOL_SET_WIDGET(Widget,inside);
static FN_GET_BOOL(Widget,take_focus);
#if defined(HAVE_FL_WIDGET_IMAGE)
static FN_GETSET_OBJECT(Widget,Fl_Image,image);
static FN_GETSET_OBJECT(Widget,Fl_Image,deimage);
#endif
#if defined(HAVE_FL_WIDGET_TOOLTIP)
static FN_GETSET_STR_COPY(Widget,tooltip);
#endif

/*
static FN_GET_INT(Widget,x);
static FN_GET_INT(Widget,y);
static FN_GET_INT(Widget,w);
static FN_GET_INT(Widget,h);
*/

static VALUE rb_FLWidget_x(int argc, VALUE argv[], VALUE self)
{
  RBFLWidget *widget = (RBFLWidget*)rb_to_fl(self);
  VALUE x;

  switch( rb_scan_args(argc, argv, "01", &x) ){
  case 0:
    return INT2NUM(widget->x());
  case 1:
    widget->resize(NUM2INT(x), widget->y(), widget->w(), widget->h());
    return x;
  };
}

static VALUE rb_FLWidget_y(int argc, VALUE argv[], VALUE self)
{
  RBFLWidget *widget = (RBFLWidget*)rb_to_fl(self);
  VALUE y;

  switch( rb_scan_args(argc, argv, "01", &y) ){
  case 0:
    return INT2NUM(widget->y());
  case 1:
    widget->resize(widget->x(), NUM2INT(y), widget->w(), widget->h());
    return y;
  };
}

static VALUE rb_FLWidget_w(int argc, VALUE argv[], VALUE self)
{
  RBFLWidget *widget = (RBFLWidget*)rb_to_fl(self);
  VALUE w;

  switch( rb_scan_args(argc, argv, "01", &w) ){
  case 0:
    return INT2NUM(widget->w());
  case 1:
    widget->resize(widget->x(), widget->y(), NUM2INT(w), widget->h());
    return w;
  };
}

static VALUE rb_FLWidget_h(int argc, VALUE argv[], VALUE self)
{
  RBFLWidget *widget = (RBFLWidget*)rb_to_fl(self);
  VALUE h;

  switch( rb_scan_args(argc, argv, "01", &h) ){
  case 0:
    return INT2NUM(widget->h());
  case 1:
    widget->resize(widget->x(), widget->y(), widget->w(), NUM2INT(h));
    return h;
  };
}

static VALUE rb_FLWidget_hash(VALUE self)
{
  RBFLWidget *widget = (RBFLWidget*)rb_to_fl(self);
  return INT2NUM((long)widget);
}

static VALUE rb_FLWidget_eql(VALUE self, VALUE other)
{
  RBFLWidget *widget1 = (RBFLWidget*)rb_to_fl(self);
  RBFLWidget *widget2 = (RBFLWidget*)rb_to_fl(other);
  if( self == other ){
    return Qtrue;
  }
  else{
    return Qfalse;
  }
}

extern "C" void
Init_flwidget(){
  rb_fltk_mark_stack = NULL;

  rb_cFLWidget = rb_define_class_under(rb_mFLTK, "Widget", rb_cObject);
  rb_fltk_append_mark(rb_cFLWidget, rb_flwidget_mark);

  DEF_S_METHOD(Widget, new, -1);
  DEF_INIT_METHOD(Widget);

  RB_DEF_METHOD(rb_cFLWidget, "hash", rb_FLWidget_hash, 0);
  RB_DEF_METHOD(rb_cFLWidget, "eql?", rb_FLWidget_eql, 1);
  RB_DEF_METHOD(rb_cFLWidget, "==", rb_FLWidget_eql, 1);

  RB_DEF_METHOD(rb_cFLWidget, "destroy", rb_fl_destroy, 0);
  RB_DEF_METHOD(rb_cFLWidget, "destroyed?", rb_fl_destroyed, 0);
  RB_DEF_METHOD(rb_cFLWidget, "valid?", rb_fl_valid, 0);
  DEF_METHOD(Widget,callback,-1);
  DEF_METHOD_EQ(Widget,user_data,-1);

  DEF_WIDGET_METHODS(Widget);

  DEF_PREFIXED_METHOD_EQ(Widget,widget,type,-1);
  DEF_METHOD(Widget,do_callback,-1);
  DEF_METHOD(Widget,window,0);
  DEF_METHOD(Widget,parent,0);
  DEF_METHOD_EQ(Widget,x,-1);
  DEF_METHOD_EQ(Widget,y,-1);
  DEF_METHOD_EQ(Widget,w,-1);
  DEF_METHOD_EQ(Widget,h,-1);
  DEF_METHOD(Widget,position,2);
  DEF_METHOD(Widget,size,2);
  DEF_METHOD_EQ(Widget,box,-1);
  DEF_METHOD_EQ(Widget,color,-1);
  rb_define_alias(rb_cFLWidget,"widget_color","color");
  DEF_METHOD_EQ(Widget,selection_color,-1);
  DEF_METHOD_EQ(Widget,label,-1);
  DEF_METHOD_EQ(Widget,labeltype,-1);
  DEF_METHOD_EQ(Widget,labelfont,-1);
  DEF_METHOD_EQ(Widget,labelcolor,-1);
  DEF_METHOD_EQ(Widget,labelsize,-1);
  DEF_METHOD(Widget,show,0);
  DEF_METHOD(Widget,hide,0);
  rb_define_alias(rb_cFLWidget,"widget_show","show");
  rb_define_alias(rb_cFLWidget,"widget_hide","hide");
  DEF_METHOD(Widget,redraw,0);
  DEF_METHOD_EQ(Widget,when,-1);
  DEF_METHOD_EQ(Widget,align,-1);
  DEF_METHOD_Q(Widget,visible,0);
  DEF_METHOD_Q(Widget,visible_r,0);
  DEF_METHOD_Q(Widget,active,0);
  DEF_METHOD_Q(Widget,active_r,0);
  DEF_METHOD(Widget,activate,0);
  DEF_METHOD(Widget,deactivate,0);
  DEF_METHOD_Q(Widget,output,0);
  DEF_METHOD(Widget,set_output,0);
  DEF_METHOD(Widget,clear_output,0);
  DEF_METHOD_Q(Widget,takesevents,0);
  DEF_METHOD_Q(Widget,damage,-1);
  DEF_METHOD_Q(Widget,contains,1);
  DEF_METHOD_Q(Widget,inside,1);
  DEF_METHOD_Q(Widget,take_focus,0);
#if defined(HAVE_FL_WIDGET_IMAGE)
  DEF_METHOD_EQ(Widget, image, -1);
  DEF_METHOD_EQ(Widget, deimage, -1);
#endif
#if defined(HAVE_FL_WIDGET_TOOLTIP)
  DEF_METHOD_EQ(Widget, tooltip, -1);
#endif
}
