# $Id: fltk.rb,v 1.13 2006/12/31 11:58:23 jeremy_henty Exp $

module FLTK

  @have_kernel_warn =
    Kernel.respond_to?(:warn)
  @warned = Hash.new { false }

  class << self

    def warn_once(key,message)
      return unless 
        @have_kernel_warn &&
        !@warned[key]
      Kernel.warn("Ruby-FLTK warning: #{message}")
      @warned[key] = true 
    end

    def deprecated(key,bad,good)
      message =
        "#{bad} is deprecated, use #{good}"
      warn_once(key,message)
    end

  end

  module_function

  # wrapped methods 

  # wrapped methods 1: push_clip and push_no_clip take an optional
  # block

  def push_clip(*args)
    push_clip_(*args)
    if block_given?
      begin yield
      ensure pop_clip
      end
    end
  end

  def push_no_clip(*args)
    push_no_clip_(*args)
    if block_given?
      begin yield
      ensure pop_clip
      end
    end
  end

  # wrapped methods 2: color takes an optional block, restoring the
  # original color after yielding.

  def color(*args)
    if block_given?
      _color = color_
      color_(*args) 
      begin yield
      ensure color_(_color)
      end
    else color_(*args) 
    end
  end

  def color=(_color) ; color_(_color) ; end

  # wrapped methods 3: line_style takes an optional block, restoring
  # the default after yielding.

  def line_style(*args)
    if block_given?
      line_style_(*args)
      begin yield
      ensure line_style_(0)
      end
    else
      line_style_(*args)
    end
  end

  def line_style=(_line_style)
    line_style_(_line_style)
  end

  # widget convenience methods 

  class Widget

    # apply a FLTK method to the widget's rectangle 

    # drawing rectangles 
    def rectf ; FLTK.rectf(x,y,w,h) ; end
    def rect  ; FLTK.rect(x,y,w,h)  ; end

    # clipping 
    def push_clip(&block) ; FLTK.push_clip(x,y,w,h,&block) ; end
    def pop_clip ; FLTK.pop_clip ; end
    def clip(&block)
      FLTK.deprecated(:fltk_widget_clip,
                      "FLTK::Widget#clip",
                      "FLTK::Widget#push_clip")
      push_clip(&block)
    end

  end

  # color classes 

  module Color

    module Common
      INSPECT_FORMAT =
        "#<%s:%s>"
      def inspect
        INSPECT_FORMAT %
          [ self.class, self, ]
      end
    end

    class Index

      include Common

      def rgb ; RGB.new(r,g,b) ; end
      def rgb_dup ; rgb ; end

      TO_S_FORMAT =
        "#0x%02X"

      def to_s
        TO_S_FORMAT %
          [ index, ]
      end

    end

    class RGB

      include Common

      def rgb ; self ; end
      def rgb_dup ; dup ; end

      TO_S_FORMAT =
        "#%02X%02X%02X"
      def to_s
        TO_S_FORMAT %
          [ r, g, b, ]
      end

    end

  end

  # deprecated methods 

  def ask0(arg)
    FLTK.deprecated(:fltk_ask0,"FLTK.ask0","FLTK.choice0")
    ask0_(arg)
  end

  def ask(*args)
    FLTK.deprecated(:fltk_ask,"FLTK.ask","FLTK.choice")
    ask_(*args)
  end

  def clip(*args,&block)
    FLTK.deprecated(:fltk_clip,"FLTK.clip","FLTK.push_clip")
    push_clip(*args,&block)
  end

end
