# $Id: sample8.rb,v 1.11 2006/12/31 11:58:23 jeremy_henty Exp $

require 'fltk'

module Browser

  def initialize(display_cb,legend,items)
    super(100,300)
    items.each{|item| add(item) }
    callback { display_cb.call(legend,select_list) }
  end

  # SelectBrowser changes this 
  def select_list
    (1..size_).select {|item| selected_(item) }
  end

  # CheckBrowser changes these 
  def size_ ; size ; end
  def selected_(item) ; selected(item) ; end

end

class ReadOnlyBrowser < FLTK::Browser
  include Browser
end

class SelectBrowser < FLTK::SelectBrowser
  include Browser
  def select_list ; [ value, ] ; end
end

class HoldBrowser < FLTK::HoldBrowser
  include Browser
end

class MultiBrowser < FLTK::MultiBrowser
  include Browser
end

class CheckBrowser < FLTK::CheckBrowser
  include Browser
  alias size_ nitems
  alias selected_ checked

  # FLTK::CheckBrowser doesn't do the callback when clicked(!), so we
  # do it ourselves.

  def handle(event)
    super
  ensure
    case event
    when FLTK::PUSH : do_callback
    end
  end

end

browsers =
  [ [ ReadOnlyBrowser, 'read-only', 
      'Read', 'only', ], 
    [ HoldBrowser, 'persistent', 
      'Persistent', 'single', 'selection', ], 
    [ SelectBrowser, 'transient', 
      'Transient', 'single', 'selection', ], 
    [ MultiBrowser, 'multiple', 
      'Multiple', 'selections', 'with',
      'ctrl-click,', 'shift-click,',
      'and', 'drag', ], 
    [ CheckBrowser, 'checked', 
      'Multiple', 'checked', 'selections', ], 
  ]

# the display callback 
message = nil # will be defined later 
display_cb = Proc.new{|legend,items|
  text = items.empty? ? "<None>" : items.join(", ")
  message.label("Selection (#{legend}): #{text}")
}

width = browsers.size*100
FLTK::Window.new(width,230,'Lists'){|win|
  FLTK::Pack.new(win.w,win.h){|pack|
    pack.pack_type(FLTK::VERTICAL)
    FLTK::Pack.new(win.w,200){|pack|
      pack.pack_type(FLTK::HORIZONTAL)
      browsers.each {|_class,legend,*items|
        _class.new(display_cb,legend,items)
      }
    }
    message = FLTK::Box.new(win.w,30)
    display_cb.call("initial",[])
  }
  win.show
}

FLTK.run
