# $Id: password.rb,v 1.3 2006/12/31 11:58:24 jeremy_henty Exp $

# This program is a part of the Ruby/FLTK tests.
# Author:: Jeremy Henty <mailto:jeremy_henty@users.sourceforge.net>
# Copyright:: Copyright (c) 2006 Jeremy Henty 
# License:: LGPL

=begin

Demonstrate FLTK.password (which passes its arguments to sprintf) and
FLTK.password0 (which does not).

=end

require 'fltk'

stuff =
  [ [ :password0, ], 
    [ :password0, "This is not a format, even though it contains a '%'!", ], 
    [ :password0, "Password with a default", "whatever", ], 
    [ :password0, "Default set to nil", nil, ], 
    [ :password0, "Too many arguments!", "whatever", "what's this", ], 
    [ :password, ],  
    [ :password, 
      "This string is a format with an escaped '%%'!", ], 
    [ :password, 
      "Format (note escaped '%%') and default value", 
      "whatever", ], 
    [ :password, 
      "Format (note escaped '%%') and nil default", 
      nil, ], 
    [ :password, 
      "This format interpolates the string '%s' and the number %d.", 
      nil, "Grand Wazoo", 42, ], 
    [ :password, 
      "Formatted %s with %s", 
      "whatever", "string", "default value", ], 
    [ :password, 
      "A '%S' makes this string a bad format, " +
      "so it will raise an error.", ], 
    [ :password,
      "Too few format arguments (%s %s).", "hmm?", ], 
  ]

button_height = 30
win_height = stuff.size * button_height
FLTK::Window.new(900,win_height,"Password"){|win|
  FLTK::Pack.new(win.w,1){|pack|
    pack.pack_type(FLTK::VERTICAL)
    stuff.each{|_m,*_a|
      label = "FLTK.%s(%s)" %
      [ _m, _a.collect{|_a0| _a0.inspect }.join(","), ]
      answers = _a[1..3]
      button = FLTK::Button.new(0,button_height,label){
        begin FLTK.message("Answer: %s",FLTK.send(_m,*_a))
        rescue => err ; FLTK.alert("Error: %s",err)
        end
      }
      button.align(FLTK::ALIGN_INSIDE |
                   FLTK::ALIGN_LEFT   )
    }
  }
  win.show
}

FLTK.run
